'''
Functions and classes to work with symbolic atoms.

Examples
--------

    >>> from clingo.symbol import Function, Number
    >>> from clingo.control import Control
    >>> ctl = Control()
    >>> ctl.add('base', [], """\\
    ... p(1).
    ... { p(3) }.
    ... #external p(1..3).
    ...
    ... q(X) :- p(X).
    ... """)
    >>> ctl.ground([("base", [])])
    >>> len(ctl.symbolic_atoms)
    6
    >>> ctl.symbolic_atoms[Function("p", [Number(2)])] is not None
    True
    >>> ctl.symbolic_atoms[Function("p", [Number(4)])] is None
    True
    >>> ctl.symbolic_atoms.signatures
    [('p', 1, True), ('q', 1, True)]
    >>> [(str(x.symbol), x.is_fact, x.is_external)
    ...  for x in ctl.symbolic_atoms.by_signature("p", 1)]
    [('p(1)', True, False), ('p(3)', False, False), ('p(2)', False, True)]
'''

from typing import Collection, Iterator, List, Optional, Tuple

from ._internal import _c_call, _ffi, _handle_error, _lib, _to_str
from .symbol import Symbol

__all__ = [ 'SymbolicAtom', 'SymbolicAtoms' ]

class SymbolicAtom:
    '''
    Captures a symbolic atom and provides properties to inspect its state.
    '''
    def __init__(self, rep, it):
        self._rep = rep
        self._it = it

    def match(self, name: str, arity: int, positive: bool=True) -> bool:
        '''
        Check if the atom matches the given signature.

        Parameters
        ----------
        name
            The name of the function.

        arity
            The arity of the function.

        positive
            Whether to match positive or negative signatures.

        Returns
        -------
        Whether the atom matches.

        See Also
        --------
        clingo.symbol.Symbol.match
        '''
        return self.symbol.match(name, arity, positive)

    @property
    def is_external(self) -> bool:
        '''
        Whether the atom is an external atom.
        '''
        return _c_call('bool', _lib.clingo_symbolic_atoms_is_external, self._rep, self._it)

    @property
    def is_fact(self) -> bool:
        '''
        Whether the atom is a fact.
        '''
        return _c_call('bool', _lib.clingo_symbolic_atoms_is_fact, self._rep, self._it)

    @property
    def literal(self) -> int:
        '''
        The program literal associated with the atom.
        '''
        return _c_call('clingo_literal_t', _lib.clingo_symbolic_atoms_literal, self._rep, self._it)

    @property
    def symbol(self) -> Symbol:
        '''
        The representation of the atom in form of a symbol.
        '''
        return Symbol(_c_call('clingo_symbol_t', _lib.clingo_symbolic_atoms_symbol, self._rep, self._it))

class SymbolicAtoms(Collection[SymbolicAtom]):
    '''
    This class provides read-only access to the atom base of the grounder.
    '''
    def __init__(self, rep):
        self._rep = rep

    def _iter(self, p_sig) -> Iterator[SymbolicAtom]:
        p_it = _ffi.new('clingo_symbolic_atom_iterator_t*')
        p_valid = _ffi.new('bool*')
        _handle_error(_lib.clingo_symbolic_atoms_begin(self._rep, p_sig, p_it))
        while _c_call(p_valid, _lib.clingo_symbolic_atoms_is_valid, self._rep, p_it[0]):
            yield SymbolicAtom(self._rep, p_it[0])
            _handle_error(_lib.clingo_symbolic_atoms_next(self._rep, p_it[0], p_it))

    def __iter__(self) -> Iterator[SymbolicAtom]:
        yield from self._iter(_ffi.NULL)

    def __contains__(self, symbol) -> bool:
        if not isinstance(symbol, Symbol):
            return False

        it = _c_call('clingo_symbolic_atom_iterator_t', _lib.clingo_symbolic_atoms_find, self._rep, symbol._rep)

        return _c_call('bool', _lib.clingo_symbolic_atoms_is_valid, self._rep, it)

    def __getitem__(self, symbol: Symbol) -> Optional[SymbolicAtom]:
        it = _c_call('clingo_symbolic_atom_iterator_t', _lib.clingo_symbolic_atoms_find, self._rep, symbol._rep)

        if not _c_call('bool', _lib.clingo_symbolic_atoms_is_valid, self._rep, it):
            return None

        return SymbolicAtom(self._rep, it)

    def __len__(self) -> int:
        return _c_call('size_t', _lib.clingo_symbolic_atoms_size, self._rep)

    def by_signature(self, name: str, arity: int, positive: bool=True) -> Iterator[SymbolicAtom]:
        '''
        Return an iterator over the symbolic atoms with the given signature.

        Arguments
        ---------
        name
            The name of the signature.
        arity
            The arity of the signature.
        positive
            The sign of the signature.
        '''
        p_sig = _ffi.new('clingo_signature_t*')
        _handle_error(_lib.clingo_signature_create(name.encode(), arity, positive, p_sig))
        yield from self._iter(p_sig)

    @property
    def signatures(self) -> List[Tuple[str,int,bool]]:
        '''
        The list of predicate signatures occurring in the program.

        The Boolean indicates the sign of the signature.
        '''
        size = _c_call('size_t', _lib.clingo_symbolic_atoms_signatures_size, self._rep)

        p_sigs = _ffi.new('clingo_signature_t[]', size)
        _handle_error(_lib.clingo_symbolic_atoms_signatures(self._rep, p_sigs, size))

        return [ (_to_str(_lib.clingo_signature_name(c_sig)),
                  _lib.clingo_signature_arity(c_sig),
                  _lib.clingo_signature_is_positive(c_sig)) for c_sig in p_sigs ]



'''
Functions and classes to implement applications based on clingo.

Examples
--------
The following example reproduces the default clingo application:

    import sys
    from clingo.application import Application, clingo_main

    class ClingoApp(Application):
        def __init__(self, name):
            self.program_name = name

        def main(self, ctl, files):
            for f in files:
                ctl.load(f)
            if not files:
                ctl.load("-")
            ctl.ground([("base", [])])
            ctl.solve()

    clingo_main(ClingoApp(sys.argv[0]), sys.argv[1:])
'''

from typing import Any, Callable, List, Optional, Sequence
from abc import ABCMeta, abstractmethod
import sys

from ._internal import _cb_error_print, _cb_error_panic, _ffi, _handle_error, _lib, _overwritten, _to_str
from .core import MessageCode
from .solving import Model
from .control import Control

__all__ = [ 'Application', 'ApplicationOptions', 'Flag', 'clingo_main' ]

class Flag:
    '''
    Helper object to parse command-line flags.

    Parameters
    ----------
    value
        The initial value of the flag.
    '''
    def __init__(self, value: bool=False):
        self._flag = _ffi.new('bool*', value)

    def __bool__(self):
        return self.flag

    @property
    def flag(self) -> bool:
        '''
        The value of the flag.
        '''
        return self._flag[0]

    @flag.setter
    def flag(self, value: bool):
        self._flag[0] = value

class ApplicationOptions(metaclass=ABCMeta):
    '''
    Object to add custom options to a clingo based application.
    '''
    def __init__(self, rep, mem):
        self._rep = rep
        self._mem = mem

    def add(self, group: str, option: str, description: str, parser: Callable[[str], bool],
            multi: bool=False, argument: Optional[str]=None) -> None:
        '''
        Add an option that is processed with a custom parser.

        Parameters
        ----------
        group
            Options are grouped into sections as given by this string.
        option
            Parameter option specifies the name(s) of the option. For example,
            `"ping,p"` adds the short option `-p` and its long form `--ping`. It is
            also possible to associate an option with a help level by adding `",@l"` to
            the option specification. Options with a level greater than zero are only
            shown if the argument to help is greater or equal to `l`.
        description
            The description of the option shown in the help output.
        parser
            An option parser is a function that takes a string as input and returns
            true or false depending on whether the option was parsed successively.
        multi
            Whether the option can appear multiple times on the command-line.
        argument
            Optional string to change the value name in the generated help.

        Notes
        -----
        The parser callback has to take care of storing the semantic value of
        the option somewhere.
        '''
        # pylint: disable=protected-access
        c_data = _ffi.new_handle(parser)
        self._mem.append(c_data)

        _handle_error(_lib.clingo_options_add(
            self._rep, group.encode(), option.encode(), description.encode(),
            _lib.pyclingo_application_options_parse, c_data,
            multi, argument.encode() if argument is not None else _ffi.NULL))

    def add_flag(self, group: str, option: str, description: str, target: Flag) -> None:
        '''
        Add an option that is a simple flag.

        This function is similar to `ApplicationOptions.add` but simpler because
        it only supports flags, which do not have values. Note that the target
        parameter must be of type Flag, which is set to true if the flag is passed on
        the command line.

        Parameters
        ----------
        group
            Options are grouped into sections as given by this string.
        option
            Same as for `ApplicationOptions.add`.
        description
            The description of the option shown in the help output.
        target
            The object that receives the value.
        '''
        # pylint: disable=protected-access
        self._mem.append(target)
        _handle_error(_lib.clingo_options_add_flag(
            self._rep, group.encode(), option.encode(), description.encode(),
            target._flag))

@_ffi.def_extern(onerror=_cb_error_panic, name='pyclingo_application_options_parse')
def _pyclingo_application_options_parse(value, data):
    return _ffi.from_handle(data)(_to_str(value))

class Application(metaclass=ABCMeta):
    '''
    Interface that has to be implemented to customize clingo.
    '''
    program_name: str
    '''
    Program name defaulting to `'clingo'` used in the help output.
    '''
    version: str
    '''
    Version string defaulting to clingo's version.
    '''
    message_limit: int
    '''
    Maximum number of messages defaulting to `20` passed to the logger.
    '''

    @abstractmethod
    def main(self, control: Control, files: Sequence[str]) -> None:
        '''
        Function to replace clingo's default main function.

        This function must be implemented.

        Parameters
        ----------
        control
            The main control object.
        files
            The files passed to clingo_main.
        '''

    def register_options(self, options: ApplicationOptions) -> None:
        '''
        Function to register custom options.

        Parameters
        ----------
        options
            Object to register additional options
        '''

    def validate_options(self) -> bool:
        '''
        Function to validate custom options.

        Returns
        -------
        This function should return false if option validation fails.
        '''

    def print_model(self, model: Model, printer: Callable[[], None]) -> None:
        '''
        Function to print additional information when the text output is used.

        Parameters
        ----------
        model
            The current model
        printer
            The default printer as used in clingo.
        '''

    def logger(self, code: MessageCode, message: str) -> None:
        '''
        Function to intercept messages normally printed to standard error.

        By default, messages are printed to standard error.

        Parameters
        ----------
        code
            The message code.
        message
            The message string.

        Notes
        -----
        This function should not raise exceptions.
        '''

def clingo_main(application: Application, arguments: Optional[Sequence[str]]=None) -> int:
    '''
    Runs the given application using clingo's default output and signal handling.

    The application can overwrite clingo's default behaviour by registering
    additional options and overriding its default main function.

    Parameters
    ----------
    application
        The application object.
    arguments
        The command line arguments excluding the program name.

        If omitted, then `sys.argv[1:]` is used.

    Returns
    -------
    The exit code of the application.

    Notes
    -----
    The main function of the `Application` interface has to be implemented. All
    other members are optional.
    '''
    if arguments is None:
        arguments = sys.argv[1:]

    # pylint: disable=dangerous-default-value,protected-access,line-too-long
    c_application = _ffi.new('clingo_application_t*', (
        _lib.pyclingo_application_program_name if _overwritten(Application, application, "program_name") else _ffi.NULL,
        _lib.pyclingo_application_version if _overwritten(Application, application, "version") else _ffi.NULL,
        _lib.pyclingo_application_message_limit if _overwritten(Application, application, "message_limit") else _ffi.NULL,
        _lib.pyclingo_application_main if _overwritten(Application, application, "main") else _ffi.NULL,
        _lib.pyclingo_application_logger if _overwritten(Application, application, "logger") else _ffi.NULL,
        _lib.pyclingo_application_print_model if _overwritten(Application, application, "print_model") else _ffi.NULL,
        _lib.pyclingo_application_register_options if _overwritten(Application, application, "register_options") else _ffi.NULL,
        _lib.pyclingo_application_validate_options if _overwritten(Application, application, "validate_options") else _ffi.NULL))

    mem: List[Any] = []
    c_data = _ffi.new_handle((application, mem))

    return _lib.clingo_main(
        c_application,
        [ _ffi.new('char[]', arg.encode()) for arg in arguments ], len(arguments),
        c_data)

@_ffi.def_extern(onerror=_cb_error_panic, name='pyclingo_application_program_name')
def _pyclingo_application_program_name(data):
    app, mem = _ffi.from_handle(data)
    mem.append(_ffi.new('char[]', app.program_name.encode()))
    return mem[-1]

@_ffi.def_extern(onerror=_cb_error_panic, name='pyclingo_application_version')
def _pyclingo_application_version(data):
    app, mem = _ffi.from_handle(data)
    mem.append(_ffi.new('char[]', app.version.encode()))
    return mem[-1]

@_ffi.def_extern(onerror=_cb_error_panic, name='pyclingo_application_message_limit')
def _pyclingo_application_message_limit(data):
    app = _ffi.from_handle(data)[0]
    return app.message_limit

@_ffi.def_extern(onerror=_cb_error_panic, name='pyclingo_application_logger')
def _pyclingo_application_logger(code, message, data):
    app = _ffi.from_handle(data)[0]
    return app.logger(MessageCode(code), _to_str(message))

@_ffi.def_extern(onerror=_cb_error_print, name='pyclingo_application_main')
def _pyclingo_application_main(control, files, size, data):
    app = _ffi.from_handle(data)[0]
    app.main(Control(control), [ _to_str(files[i]) for i in range(size) ])
    return True

@_ffi.def_extern(onerror=_cb_error_print, name='pyclingo_application_print_model')
def _pyclingo_application_print_model(model, printer, printer_data, data):
    def py_printer():
        _handle_error(printer(printer_data))
    app = _ffi.from_handle(data)[0]
    app.print_model(Model(model), py_printer)
    return True

@_ffi.def_extern(onerror=_cb_error_panic, name='pyclingo_application_register_options')
def _pyclingo_application_register_options(options, data):
    app, mem = _ffi.from_handle(data)
    app.register_options(ApplicationOptions(options, mem))
    return True

@_ffi.def_extern(onerror=_cb_error_panic, name='pyclingo_application_validate_options')
def _pyclingo_application_validate_options(data):
    app = _ffi.from_handle(data)[0]
    return app.validate_options()



'''
Module to work with clingo's non-ground program representation.

Grammar
-------
The grammar below defines valid ASTs. For each upper case identifier there is a
matching function in the module. Arguments follow in parenthesis: each having a
type given on the right-hand side of the colon. The symbols `?`, `*`, and `+`
are used to denote optional arguments (`None` encodes abscence), list
arguments, and non-empty list arguments.

```
# Terms

term = SymbolicTerm
        ( location : Location
        , symbol   : clingo.Symbol
        )
     | Variable
        ( location : Location
        , name     : str
        )
     | UnaryOperation
        ( location      : Location
        , operator_type : UnaryOperator
        , argument      : term
        )
     | BinaryOperation
        ( location      : Location
        , operator_type : BinaryOperator
        , left          : term
        , right         : term
        )
     | Interval
        ( location : Location
        , left     : term
        , right    : term
        )
     | Function
        ( location  : Location
        , name      : str
        , arguments : term*
        , external  : bool
        )
     | Pool
        ( location  : Location
        , arguments : term*
        )

csp_term = CSPSum
            ( location : Location
            , terms    : CSPProduct
                          ( location    : Location
                          , coefficient : term
                          , variable    : term?
                          )*
            )

theory_term = SymbolicTerm
               ( location : Location
               , symbol   : clingo.Symbol
               )
            | Variable
               ( location : Location
               , name     : str
               )
            | TheorySequence
               ( location : Location
               , sequence_type : TheorySequenceType
               , terms         : theory_term*
               )
            | TheoryFunction
               ( location  : Location
               , name      : str
               , arguments : theory_term*
               )
            | TheoryUnparsedTerm
               ( location : Location
               , elements : TheoryUnparsedTermElement
                             ( operators : str*
                             , term      : theory_term
                             )+
               )

# Literals

symbolic_atom = SymbolicAtom
                 ( symbol : term
                 )

literal = Literal
           ( location : Location
           , sign     : Sign
           , atom     : Comparison
                         ( comparison : ComparisonOperator
                         , left       : term
                         , right      : term
                         )
                      | BooleanConstant
                         ( value : bool
                         )
                      | symbolic_atom
           )

        | CSPLiteral
           ( location : Location
           , term     : csp_term
           , guards   : CSPGuard
                         ( comparison : ComparisonOperator
                         , term       : csp_term
                         )+
           )

# Head and Body Literals

aggregate_guard = AggregateGuard
                   ( comparison : ComparisonOperator
                   , term       : term
                   )

conditional_literal = ConditionalLiteral
                       ( location  : Location
                       , literal   : Literal
                       , condition : Literal*
                       )

aggregate = Aggregate
             ( location    : Location
             , left_guard  : aggregate_guard?
             , elements    : conditional_literal*
             , right_guard : aggregate_guard?
             )

theory_atom = TheoryAtom
               ( location : Location
               , term     : term
               , elements : TheoryAtomElement
                             ( terms     : theory_term*
                             , condition : literal*
                             )*
               , guard    : TheoryGuard
                             ( operator_name : str
                             , term          : theory_term
                             )?
               )

body_atom = aggregate
          | BodyAggregate
             ( location    : Location
             , left_guard  : aggregate_guard?
             , function    : AggregateFunction
             , elements    : BodyAggregateElement
                              ( terms     : term*
                              , condition : literal*
                              )*
             , right_guard : aggregate_guard?
             )
          | Disjoint
             ( location : Location
             , elements : DisjointElement
                           ( location  : Location
                           , terms     : term*
                           , term      : csp_term
                           , condition : literal*
                           )*
             )
          | theory_atom

body_literal = literal
             | conditional_literal
             | Literal
                ( location : Location
                , sign     : Sign
                , atom     : body_atom
                )

head = literal
     | aggregate
     | HeadAggregate
        ( location    : Location
        , left_guard  : aggregate_guard?
        , function    : AggregateFunction
        , elements    : HeadAggregateElement
                         ( terms     : term*
                         , condition : conditional_literal
                         )*
        , right_guard : aggregate_guard?
        )
     | Disjunction
        ( location : Location
        , elements : conditional_literal*
        )
     | theory_atom

# Statements

statement = Rule
             ( location : Location
             , head     : head
             , body     : body_literal*
             )
          | Definition
             ( location   : Location
             , name       : str
             , value      : term
             , is_default : bool
             )
          | ShowSignature
             ( location   : Location
             , name       : str
             , arity      : int
             , sign       : bool
             , csp        : bool
             )
          | Defined
             ( location   : Location
             , name       : str
             , arity      : int
             , sign       : bool
             )
          | ShowTerm
             ( location : Location
             , term     : term
             , body     : body_literal*
             , csp      : bool
             )
          | Minimize
             ( location : Location
             , weight   : term
             , priority : term
             , terms    : term*
             , body     : body_literal*
             )
          | Script
             ( location : Location
             , name     : str
             , code     : str
             )
          | Program
             ( location   : Location
             , name       : str
             , parameters : Id
                             ( location : Location
                             , id       : str
                             )*
             )
          | External
             ( location : Location
             , atom     : symbolic_atom
             , body     : body_literal*
             , type     : term
             )
          | Edge
             ( location : Location
             , u        : term
             , v        : term
             , body     : body_literal*
             )
          | Heuristic
             ( location : Location
             , atom     : symbolic_atom
             , body     : body_literal*
             , bias     : term
             , priority : term
             , modifier : term
             )
          | ProjectAtom
             ( location : Location
             , atom     : symbolic_atom
             , body     : body_literal*
             )
          | ProjectSignature
             ( location : Location
             , name     : str
             , arity    : int
             , sign     : bool
             )
          | TheoryDefinition
             ( location : Location
             , name     : str
             , terms    : TheoryTermDefinition
                           ( location  : Location
                           , name      : str
                           , operators : TheoryOperatorDefinition
                                          ( location      : Location
                                          , name          : str
                                          , priority      : int
                                          , operator_type : TheoryOperatorType
                                          )*
                           )*
             , atoms    : TheoryAtomDefinition
                           ( location  : Location
                           , atom_type : TheoryAtomType
                           , name      : str
                           , arity     : int
                           , term      : str
                           , guard     : TheoryGuardDefinition
                                          ( operators : str*
                                          , term      : str
                                          )?
                           )*
             )
```

Examples
--------
The following example parses a program from a string and passes the resulting
`AST` to the builder:

    >>> from clingo import Control, ast
    >>> from clingo.ast import Location, ProgramBuilder, Position, parse_string
    >>>
    >>> ctl = Control()
    >>>
    >>> with ProgramBuilder(ctl) as bld:
    ...     # parse from string
    ...     parse_string('a.', bld.add)
    ...     # build rule manually
    ...     pos = Position('<string>', 1, 1)
    ...     loc = Location(pos, pos)
    ...     fun = ast.Function(loc, 'b', [], False)
    ...     atm = ast.SymbolicAtom(fun)
    ...     lit = ast.Literal(loc, ast.Sign.NoSign, atm)
    ...     bld.add(ast.Rule(loc, lit, []))
    ...
    >>> ctl.ground([('base', [])])
    >>> print(ctl.solve(on_model=print))
    a b
    SAT

The next example shows how to transform ASTs using the `Transformer` class:

    >>> from clingo.ast import Transformer, Variable, parse_string
    >>>
    >>> class VariableRenamer(Transformer):
    ...     def visit_Variable(self, node):
    ...         return node.update(name='_' + node.name)
    ...
    >>> vrt = VariableRenamer()
    >>> parse_string('p(X) :- q(X).', lambda stm: print(str(vrt(stm))))
    #program base.
    p(_X) :- q(_X).
'''

from enum import Enum, IntEnum
from typing import Any, Callable, ContextManager, Dict, List, MutableSequence, NamedTuple, Optional, Sequence, Tuple, Union
from collections import abc
from functools import total_ordering

from ._internal import (_CBData, _Error,
                        _cb_error_handler, _c_call, _ffi, _handle_error, _lib, _str, _to_str)
from .util import Slice, SlicedMutableSequence
from .core import Logger
from .control import Control
from .symbol import Symbol

__all__ = [ 'AST', 'ASTSequence', 'ASTType', 'ASTValue', 'Aggregate',
            'AggregateFunction', 'AggregateGuard', 'BinaryOperation',
            'BinaryOperator', 'BodyAggregate', 'BodyAggregateElement',
            'BooleanConstant', 'Comparison', 'ComparisonOperator',
            'ConditionalLiteral', 'CspGuard', 'CspLiteral', 'CspProduct',
            'CspSum', 'Defined', 'Definition', 'Disjoint', 'DisjointElement',
            'Disjunction', 'Edge', 'External', 'Function', 'HeadAggregate',
            'HeadAggregateElement', 'Heuristic', 'Id', 'Interval', 'Literal',
            'Location', 'Minimize', 'Pool', 'Position', 'Program',
            'ProgramBuilder', 'ProjectAtom', 'ProjectSignature', 'Rule',
            'Script', 'ShowSignature', 'ShowTerm', 'Sign', 'StrSequence',
            'SymbolicTerm', 'TheoryAtom', 'TheoryAtomDefinition',
            'TheoryAtomElement', 'TheoryAtomType', 'TheoryDefinition',
            'TheoryFunction', 'TheoryGuard', 'TheoryGuardDefinition',
            'TheoryOperatorDefinition', 'TheoryOperatorType', 'TheorySequence',
            'TheorySequenceType', 'TheoryTermDefinition', 'TheoryUnparsedTerm',
            'TheoryUnparsedTermElement', 'Transformer', 'UnaryOperation',
            'UnaryOperator', 'Variable', 'parse_files', 'parse_string' ]

# pylint: disable=protected-access,invalid-name,too-many-lines,too-many-ancestors

class ASTType(Enum):
    '''
    Enumeration of ast node types.
    '''
    Id = _lib.clingo_ast_type_id
    Variable = _lib.clingo_ast_type_variable
    SymbolicTerm = _lib.clingo_ast_type_symbolic_term
    UnaryOperation = _lib.clingo_ast_type_unary_operation
    BinaryOperation = _lib.clingo_ast_type_binary_operation
    Interval = _lib.clingo_ast_type_interval
    Function = _lib.clingo_ast_type_function
    Pool = _lib.clingo_ast_type_pool
    CspProduct = _lib.clingo_ast_type_csp_product
    CspSum = _lib.clingo_ast_type_csp_sum
    CspGuard = _lib.clingo_ast_type_csp_guard
    BooleanConstant = _lib.clingo_ast_type_boolean_constant
    SymbolicAtom = _lib.clingo_ast_type_symbolic_atom
    Comparison = _lib.clingo_ast_type_comparison
    CspLiteral = _lib.clingo_ast_type_csp_literal
    AggregateGuard = _lib.clingo_ast_type_aggregate_guard
    ConditionalLiteral = _lib.clingo_ast_type_conditional_literal
    Aggregate = _lib.clingo_ast_type_aggregate
    BodyAggregateElement = _lib.clingo_ast_type_body_aggregate_element
    BodyAggregate = _lib.clingo_ast_type_body_aggregate
    HeadAggregateElement = _lib.clingo_ast_type_head_aggregate_element
    HeadAggregate = _lib.clingo_ast_type_head_aggregate
    Disjunction = _lib.clingo_ast_type_disjunction
    DisjointElement = _lib.clingo_ast_type_disjoint_element
    Disjoint = _lib.clingo_ast_type_disjoint
    TheorySequence = _lib.clingo_ast_type_theory_sequence
    TheoryFunction = _lib.clingo_ast_type_theory_function
    TheoryUnparsedTermElement = _lib.clingo_ast_type_theory_unparsed_term_element
    TheoryUnparsedTerm = _lib.clingo_ast_type_theory_unparsed_term
    TheoryGuard = _lib.clingo_ast_type_theory_guard
    TheoryAtomElement = _lib.clingo_ast_type_theory_atom_element
    TheoryAtom = _lib.clingo_ast_type_theory_atom
    Literal = _lib.clingo_ast_type_literal
    TheoryOperatorDefinition = _lib.clingo_ast_type_theory_operator_definition
    TheoryTermDefinition = _lib.clingo_ast_type_theory_term_definition
    TheoryGuardDefinition = _lib.clingo_ast_type_theory_guard_definition
    TheoryAtomDefinition = _lib.clingo_ast_type_theory_atom_definition
    Rule = _lib.clingo_ast_type_rule
    Definition = _lib.clingo_ast_type_definition
    ShowSignature = _lib.clingo_ast_type_show_signature
    ShowTerm = _lib.clingo_ast_type_show_term
    Minimize = _lib.clingo_ast_type_minimize
    Script = _lib.clingo_ast_type_script
    Program = _lib.clingo_ast_type_program
    External = _lib.clingo_ast_type_external
    Edge = _lib.clingo_ast_type_edge
    Heuristic = _lib.clingo_ast_type_heuristic
    ProjectAtom = _lib.clingo_ast_type_project_atom
    ProjectSignature = _lib.clingo_ast_type_project_signature
    Defined = _lib.clingo_ast_type_defined
    TheoryDefinition = _lib.clingo_ast_type_theory_definition

class AggregateFunction(IntEnum):
    '''
    Enumeration of aggegate functions.
    '''
    Count = _lib.clingo_ast_aggregate_function_count
    '''
    The `#count` function.
    '''
    Max = _lib.clingo_ast_aggregate_function_max
    '''
    The `#max` function.
    '''
    Min = _lib.clingo_ast_aggregate_function_min
    '''
    The `#min` function.
    '''
    Sum = _lib.clingo_ast_aggregate_function_sum
    '''
    The `#sum` function.
    '''
    SumPlus = _lib.clingo_ast_aggregate_function_sump
    '''
    The `#sum+` function.
    '''

class BinaryOperator(IntEnum):
    '''
    Enumeration of binary operators.
    '''
    And = _lib.clingo_ast_binary_operator_and
    '''
    For bitwise and.
    '''
    Division = _lib.clingo_ast_binary_operator_division
    '''
    For arithmetic division.
    '''
    Minus = _lib.clingo_ast_binary_operator_minus
    '''
    For arithmetic subtraction.
    '''
    Modulo = _lib.clingo_ast_binary_operator_modulo
    '''
    For arithmetic modulo.
    '''
    Multiplication = _lib.clingo_ast_binary_operator_multiplication
    '''
    For arithmetic multipilcation.
    '''
    Or = _lib.clingo_ast_binary_operator_or
    '''
    For bitwise or.
    '''
    Plus = _lib.clingo_ast_binary_operator_plus
    '''
    For arithmetic addition.
    '''
    Power = _lib.clingo_ast_binary_operator_power
    '''
    For arithmetic exponentiation.
    '''
    XOr = _lib.clingo_ast_binary_operator_xor
    '''
    For bitwise exclusive or.
    '''

class ComparisonOperator(IntEnum):
    '''
    Enumeration of comparison operators.
    '''
    Equal = _lib.clingo_ast_comparison_operator_equal
    '''
    The `=` operator
    '''
    GreaterEqual = _lib.clingo_ast_comparison_operator_greater_equal
    '''
    The `>=` operator.
    '''
    GreaterThan = _lib.clingo_ast_comparison_operator_greater_than
    '''
    The `>` operator.
    '''
    LessEqual = _lib.clingo_ast_comparison_operator_less_equal
    '''
    The `<=` operator.
    '''
    LessThan = _lib.clingo_ast_comparison_operator_less_than
    '''
    The `<` operator.
    '''
    NotEqual = _lib.clingo_ast_comparison_operator_not_equal
    '''
    The `!=` operator.
    '''

class Sign(IntEnum):
    '''
    Enumeration of signs for literals.
    '''
    DoubleNegation = _lib.clingo_ast_sign_double_negation
    '''
    For double negated literals (with prefix `not not`)
    '''
    Negation = _lib.clingo_ast_sign_negation
    '''
    For negative literals (with prefix `not`).
    '''
    NoSign = _lib.clingo_ast_sign_no_sign
    '''
    For positive literals.
    '''

class TheoryAtomType(IntEnum):
    '''
    Enumeration of theory atom types.
    '''
    Any = _lib.clingo_ast_theory_atom_definition_type_any
    '''
    For atoms that can occur anywhere in a rule.
    '''
    Body = _lib.clingo_ast_theory_atom_definition_type_body
    '''
    For atoms that can only occur in rule bodies.
    '''
    Directive = _lib.clingo_ast_theory_atom_definition_type_directive
    '''
    For atoms that can only occur in facts.
    '''
    Head = _lib.clingo_ast_theory_atom_definition_type_head
    '''
    For atoms that can only occur in rule heads.
    '''

class TheoryOperatorType(IntEnum):
    '''
    Enumeration of operator types.
    '''
    BinaryLeft = _lib.clingo_ast_theory_operator_type_binary_left
    '''
    For binary left associative operators.
    '''
    BinaryRight = _lib.clingo_ast_theory_operator_type_binary_right
    '''
    For binary right associative operator.
    '''
    Unary = _lib.clingo_ast_theory_operator_type_unary
    '''
    For unary operators.
    '''

class TheorySequenceType(IntEnum):
    '''
    Enumeration of theory term sequence types.
    '''
    List = _lib.clingo_ast_theory_sequence_type_list
    '''
    For sequences enclosed in brackets.
    '''
    Set = _lib.clingo_ast_theory_sequence_type_set
    '''
    For sequences enclosed in braces.
    '''
    Tuple = _lib.clingo_ast_theory_sequence_type_tuple
    '''
    For sequences enclosed in parenthesis.
    '''

class UnaryOperator(IntEnum):
    '''
    Enumeration of signs for literals.
    '''
    Absolute = _lib.clingo_ast_unary_operator_absolute
    '''
    For taking the absolute value.
    '''
    Minus = _lib.clingo_ast_unary_operator_minus
    '''
    For unary minus and classical negation.
    '''
    Negation = _lib.clingo_ast_unary_operator_negation
    '''
    For bitwise negation.
    '''

class ASTSequence(abc.MutableSequence):
    '''
    A sequence holding `AST` nodes.
    '''
    def __init__(self, rep, attribute):
        self._rep = rep
        self._attribute = attribute
        _lib.clingo_ast_acquire(self._rep)

    def __del__(self):
        _lib.clingo_ast_release(self._rep)

    def __len__(self) -> int:
        return _c_call('size_t', _lib.clingo_ast_attribute_size_ast_array, self._rep, self._attribute)

    def __getitem__(self, index):
        if isinstance(index, slice):
            return SlicedMutableSequence(self, Slice(index))
        size = len(self)
        if index < 0:
            index += size
        if index < 0 or index >= size:
            raise IndexError('invalid index')
        return AST(_c_call('clingo_ast_t*', _lib.clingo_ast_attribute_get_ast_at, self._rep, self._attribute, index))

    def __iter__(self):
        for index in range(len(self)):
            yield AST(_c_call('clingo_ast_t*', _lib.clingo_ast_attribute_get_ast_at, self._rep, self._attribute, index))

    def __setitem__(self, index, ast):
        if isinstance(index, slice):
            raise TypeError('slicing not implemented')
        _handle_error(_lib.clingo_ast_attribute_set_ast_at(self._rep, self._attribute, index, ast._rep))

    def __delitem__(self, index):
        if isinstance(index, slice):
            raise TypeError('slicing not implemented')
        size = len(self)
        if index < 0:
            index += size
        if index < 0 or index >= size:
            raise IndexError('invalid index')
        _handle_error(_lib.clingo_ast_attribute_delete_ast_at(self._rep, self._attribute, index))

    def insert(self, index, value):
        _handle_error(_lib.clingo_ast_attribute_insert_ast_at(self._rep, self._attribute, index, value._rep))

    def clear(self):
        '''
        Remove all elements from the sequence.
        '''
        for i in range(len(self), 0, -1):
            del self[i - 1]

    def __str__(self):
        return str(list(self))

    def __repr__(self):
        return repr(list(self))

class StrSequence(abc.MutableSequence):
    '''
    A sequence holding strings.
    '''
    def __init__(self, rep, attribute):
        self._attribute = attribute
        self._rep = rep
        _lib.clingo_ast_acquire(self._rep)

    def __del__(self):
        _lib.clingo_ast_release(self._rep)

    def __len__(self) -> int:
        return _c_call('size_t', _lib.clingo_ast_attribute_size_string_array, self._rep, self._attribute)

    def __getitem__(self, index):
        if isinstance(index, slice):
            return SlicedMutableSequence(self, Slice(index))
        size = len(self)
        if index < 0:
            index += size
        if index < 0 or index >= size:
            raise IndexError('invalid index')
        return _to_str(_c_call('char*', _lib.clingo_ast_attribute_get_string_at, self._rep, self._attribute, index))

    def __iter__(self):
        for index in range(len(self)):
            yield _to_str(_c_call('char*', _lib.clingo_ast_attribute_get_string_at, self._rep, self._attribute, index))

    def __setitem__(self, index, value):
        if isinstance(index, slice):
            raise TypeError('slicing not implemented')
        _handle_error(_lib.clingo_str_attribute_set_string_at(self._rep, self._attribute, index, value.encode()))

    def __delitem__(self, index):
        if isinstance(index, slice):
            raise TypeError('slicing not implemented')
        size = len(self)
        if index < 0:
            index += size
        if index < 0 or index >= size:
            raise IndexError('invalid index')
        _handle_error(_lib.clingo_ast_attribute_delete_string_at(self._rep, self._attribute, index))

    def insert(self, index, value):
        _handle_error(_lib.clingo_ast_attribute_insert_string_at(self._rep, self._attribute, index, value.encode()))

    def clear(self):
        '''
        Remove all elements from the sequence.
        '''
        for i in range(len(self), 0, -1):
            del self[i - 1]

    def __str__(self):
        return str(list(self))

    def __repr__(self):
        return repr(list(self))

class Position(NamedTuple):
    '''
    Class to point to a position in a text file.
    '''
    filename: str
    '''
    The file name.
    '''
    line: int
    '''
    The line number in the file.
    '''
    column: int
    '''
    The column number in the line.
    '''

class Location(NamedTuple):
    '''
    Class to point to a range in a text file.
    '''
    begin: Position
    '''
    The beginning of the range.
    '''
    end: Position
    '''
    The end of the range.
    '''

def _c_location(loc: Location):
    mema = _ffi.new('char[]', loc.begin.filename.encode())
    memb = _ffi.new('char[]', loc.end.filename.encode())
    return (_ffi.new('clingo_location_t*', (mema,
                                            memb,
                                            loc.begin.line,
                                            loc.end.line,
                                            loc.begin.column,
                                            loc.end.column)), mema, memb)

def _py_location(rep):
    return Location(
        Position(_to_str(rep.begin_file), rep.begin_line, rep.begin_column),
        Position(_to_str(rep.end_file), rep.end_line, rep.end_column))

_attribute_names = { _to_str(_lib.g_clingo_ast_attribute_names.names[i]): i
                     for i in range(_lib.g_clingo_ast_attribute_names.size) }

ASTValue = Union[str, int, Symbol, Location, None, 'AST', StrSequence, ASTSequence]
ASTUpdate = Union[str, int, Symbol, Location, None, 'AST', Sequence[str], Sequence['AST']]

@total_ordering
class AST:
    '''
    Represents a node in the abstract syntax tree.

    The attributes of an `AST` are tied to its type. They correspond to the
    grammar in the description of the `clingo.ast` module. `AST` nodes can be
    constructed using one of the functions provided in this module.

    Furthermore, AST nodes implement Python's rich comparison operators and are
    ordered structurally ignoring the location. They can also be used as
    dictionary keys. Their string representation corresponds to their gringo
    representation. In fact, the string representation of any AST obtained from
    `parse_files` and `parse_string` can be parsed again. Note that it is
    possible to construct ASTs that are not parsable, though.
    '''
    def __init__(self, rep):
        super().__setattr__('_rep', rep)

    def __eq__(self, other):
        if not isinstance(other, AST):
            return NotImplemented
        return _lib.clingo_ast_equal(self._rep, other._rep)

    def __lt__(self, other):
        if not isinstance(other, AST):
            return NotImplemented
        return _lib.clingo_ast_less_than(self._rep, other._rep)

    def __hash__(self):
        return _lib.clingo_ast_hash(self._rep)

    def __del__(self):
        _lib.clingo_ast_release(self._rep)

    def __getattr__(self, name):
        attr_id = _attribute_names.get(name)
        if attr_id is None or not _c_call('bool', _lib.clingo_ast_has_attribute, self._rep, attr_id):
            raise AttributeError(f'no attribute: {name}')
        attr_type = _c_call('clingo_ast_attribute_type_t', _lib.clingo_ast_attribute_type, self._rep, attr_id)
        if attr_type == _lib.clingo_ast_attribute_type_string:
            return _to_str(_c_call('char*', _lib.clingo_ast_attribute_get_string, self._rep, attr_id))
        if attr_type == _lib.clingo_ast_attribute_type_number:
            return _c_call('int', _lib.clingo_ast_attribute_get_number, self._rep, attr_id)
        if attr_type == _lib.clingo_ast_attribute_type_symbol:
            return Symbol(_c_call('clingo_symbol_t', _lib.clingo_ast_attribute_get_symbol,
                                                self._rep, attr_id))
        if attr_type == _lib.clingo_ast_attribute_type_location:
            return _py_location(_c_call('clingo_location_t', _lib.clingo_ast_attribute_get_location,
                                        self._rep, attr_id))
        if attr_type == _lib.clingo_ast_attribute_type_optional_ast:
            rep = _c_call('clingo_ast_t*', _lib.clingo_ast_attribute_get_optional_ast, self._rep, attr_id)
            return AST(rep) if rep != _ffi.NULL else None
        if attr_type == _lib.clingo_ast_attribute_type_ast:
            return AST(_c_call('clingo_ast_t*', _lib.clingo_ast_attribute_get_ast, self._rep, attr_id))
        if attr_type == _lib.clingo_ast_attribute_type_string_array:
            return StrSequence(self._rep, attr_id)
        assert attr_type == _lib.clingo_ast_attribute_type_ast_array
        return ASTSequence(self._rep, attr_id)

    def __setattr__(self, name, value):
        attr_id = getattr(_lib, f'clingo_ast_attribute_{name}')
        if not _c_call('bool', _lib.clingo_ast_has_attribute, self._rep, attr_id):
            raise AttributeError(f'no attribute: {name}')
        attr_type = _c_call('clingo_ast_attribute_type_t', _lib.clingo_ast_attribute_type, self._rep, attr_id)
        if attr_type == _lib.clingo_ast_attribute_type_string:
            _handle_error(_lib.clingo_ast_attribute_set_string(self._rep, attr_id, value.encode()))
        elif attr_type == _lib.clingo_ast_attribute_type_number:
            _handle_error(_lib.clingo_ast_attribute_set_number(self._rep, attr_id, value))
        elif attr_type == _lib.clingo_ast_attribute_type_symbol:
            _handle_error(_lib.clingo_ast_attribute_set_symbol(self._rep, attr_id, value._rep))
        elif attr_type == _lib.clingo_ast_attribute_type_location:
            c_loc = _c_location(value)
            _handle_error(_lib.clingo_ast_attribute_set_location(self._rep, attr_id, c_loc[0]))
        elif attr_type == _lib.clingo_ast_attribute_type_optional_ast:
            _handle_error(_lib.clingo_ast_attribute_set_optional_ast(self._rep, attr_id,
                          _ffi.NULL if value is None else value._rep))
        elif attr_type == _lib.clingo_ast_attribute_type_ast:
            _handle_error(_lib.clingo_ast_attribute_set_ast(self._rep, attr_id, value._rep))
        elif attr_type == _lib.clingo_ast_attribute_type_string_array:
            if isinstance(value, StrSequence):
                if attr_id == value._attribute and self._rep == value._rep:
                    value = list(value)
            elif not isinstance(value, list):
                value = list(value)
            str_seq = StrSequence(self._rep, attr_id)
            str_seq.clear()
            str_seq.extend(value)
        else:
            assert attr_type == _lib.clingo_ast_attribute_type_ast_array
            if isinstance(value, ASTSequence):
                if attr_id == value._attribute and self._rep == value._rep:
                    value = list(value)
            elif not isinstance(value, list):
                value = list(value)
            ast_seq = ASTSequence(self._rep, attr_id)
            ast_seq.clear()
            ast_seq.extend(value)

    def __str__(self):
        return _str(_lib.clingo_ast_to_string_size, _lib.clingo_ast_to_string, self._rep)

    def __repr__(self):
        name = str(self.ast_type).replace('ASTType', 'ast')
        args = ', '.join(repr(x) for x in self.values())
        return f'{name}({args})'

    def __copy__(self) -> 'AST':
        '''
        Return a shallow copy of the ast.
        '''
        return AST(_c_call('clingo_ast_t*', _lib.clingo_ast_copy, self._rep))

    def __deepcopy__(self, memo) -> 'AST':
        '''
        Return a deep copy of the ast.
        '''
        return AST(_c_call('clingo_ast_t*', _lib.clingo_ast_deep_copy, self._rep))

    def update(self, **kwargs: ASTUpdate) -> 'AST':
        '''
        Return a copy of the AST also updating the given attributes.

        Note that this function returns a reference to self if no arguments are
        given.
        '''
        if not kwargs:
            return self

        args = []
        for key in self.keys():
            if key in kwargs:
                args.append(kwargs[key])
            else:
                args.append(getattr(self, key))

        cons = globals()[str(self.ast_type).replace('ASTType.', '')]
        return cons(*args)

    def items(self) -> List[Tuple[str, ASTValue]]:
        '''
        The list of items of the AST node.
        '''
        return [ (name, getattr(self, name)) for name in self.keys() ]

    def keys(self) -> List[str]:
        '''
        The list of keys of the AST node.
        '''
        cons = _lib.g_clingo_ast_constructors.constructors[self.ast_type.value]
        names = _lib.g_clingo_ast_attribute_names.names
        return [ _to_str(names[cons.arguments[j].attribute]) for j in range(cons.size) ]

    def values(self) -> List[ASTValue]:
        '''
        The list of values of the AST node.
        '''
        return [ (getattr(self, name)) for name in self.keys() ]

    @property
    def ast_type(self) -> ASTType:
        '''
        The type of the node.
        '''
        return ASTType(_c_call('clingo_ast_type_t', _lib.clingo_ast_get_type, self._rep))

    @property
    def child_keys(self) -> List[str]:
        '''
        List of attribute names containing ASTs.
        '''
        cons = _lib.g_clingo_ast_constructors.constructors[self.ast_type.value]
        names = _lib.g_clingo_ast_attribute_names.names
        return [ _to_str(names[cons.arguments[j].attribute])
                 for j in range(cons.size)
                 if cons.arguments[j].type in (_lib.clingo_ast_attribute_type_ast,
                                               _lib.clingo_ast_attribute_type_optional_ast,
                                               _lib.clingo_ast_attribute_type_ast_array) ]

    def unpool(self, other: bool=True, condition: bool=True) -> List['AST']:
        '''
        Unpool the AST returning a list of ASTs without pool terms.

        Parameters
        ----------
        other
            Remove all pools except those in conditions of conditional
            literals.
        condition
            Only remove pools from conditions of conditional literals.
        '''
        unpool_type = 0
        if other:
            unpool_type |= _lib.clingo_ast_unpool_type_other
        if condition:
            unpool_type |= _lib.clingo_ast_unpool_type_condition

        ret: List[AST] = []
        error = _Error()
        cb_data = _CBData(ret.append, error)
        c_cb_data = _ffi.new_handle(cb_data)
        _handle_error(_lib.clingo_ast_unpool(self._rep, unpool_type, _lib.pyclingo_ast_callback, c_cb_data))
        return ret

@_ffi.def_extern(onerror=_cb_error_handler('data'), name='pyclingo_ast_callback')
def _pyclingo_ast_callback(ast, data):
    '''
    Low-level ast callback.
    '''
    callback = _ffi.from_handle(data).data
    _lib.clingo_ast_acquire(ast)
    callback(AST(ast))

    return True

def parse_files(files: Sequence[str], callback: Callable[[AST], None],
                logger: Optional[Logger]=None, message_limit: int=20) -> None:
    '''
    Parse the programs in the given files and return an abstract syntax tree for
    each statement via a callback.

    The function follows clingo's handling of files on the command line. Filename
    `"-"` is treated as stdin and if an empty list is given, then the parser will
    read from stdin.

    Parameters
    ----------
    files
        List of file names.
    callback
        Callable taking an ast as argument.
    logger
        Function to intercept messages normally printed to standard error.
    message_limit
        The maximum number of messages passed to the logger.

    See Also
    --------
    ProgramBuilder
    '''
    if logger is not None:
        c_logger_data = _ffi.new_handle(logger)
        c_logger = _lib.pyclingo_logger_callback
    else:
        c_logger_data = _ffi.NULL
        c_logger = _ffi.NULL

    error = _Error()
    cb_data = _CBData(callback, error)
    c_cb_data = _ffi.new_handle(cb_data)

    _handle_error(_lib.clingo_ast_parse_files([ _ffi.new('char[]', f.encode()) for f in files ], len(files),
                                              _lib.pyclingo_ast_callback, c_cb_data,
                                              c_logger, c_logger_data,
                                              message_limit), cb_data)

def parse_string(program: str, callback: Callable[[AST], None],
                 logger: Optional[Logger]=None, message_limit: int=20) -> None:
    '''
    Parse the given program and return an abstract syntax tree for each
    statement via a callback.

    Parameters
    ----------
    program
        String representation of the program.
    callback
        Callable taking an ast as argument.
    logger
        Function to intercept messages normally printed to standard error.
    message_limit
        The maximum number of messages passed to the logger.

    See Also
    --------
    ProgramBuilder
    '''
    if logger is not None:
        c_logger_data = _ffi.new_handle(logger)
        c_logger = _lib.pyclingo_logger_callback
    else:
        c_logger_data = _ffi.NULL
        c_logger = _ffi.NULL

    error = _Error()
    cb_data = _CBData(callback, error)
    c_cb_data = _ffi.new_handle(cb_data)

    _handle_error(_lib.clingo_ast_parse_string(program.encode(),
                                               _lib.pyclingo_ast_callback, c_cb_data,
                                               c_logger, c_logger_data,
                                               message_limit), cb_data)

class ProgramBuilder(ContextManager['ProgramBuilder']):
    '''
    Object to build non-ground programs.

    Parameters
    ----------
    control
        The `clingo.control.Control` object to attach the builder to.

    See Also
    --------
    parse_string, parse_files

    Notes
    -----
    This class is a context manager and must be used with Python's `with`
    statement.
    '''
    def __init__(self, control: Control):
        self._rep = _c_call('clingo_program_builder_t*', _lib.clingo_control_program_builder, control._rep)

    def __enter__(self):
        _handle_error(_lib.clingo_program_builder_begin(self._rep))
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        _handle_error(_lib.clingo_program_builder_end(self._rep))
        return False

    def add(self, statement: AST) -> None:
        '''
        Adds a statement in form of an `AST` node to the program.

        Parameters
        ----------
        statement
            The statement to add.
        '''
        _handle_error(_lib.clingo_program_builder_add(self._rep, statement._rep))

class Transformer:
    '''
    Utility class to transform ASTs.

    Classes should inherit from this class and implement functions with name
    `visit_<ast_type>` where `<ast_type>` is the type of the ASTs to visit and
    modify. Such a function should return an updated AST or the same AST if no
    change is necessary. The transformer will take care to copy all parent ASTs
    involving a modified child. Note that the class works like a visitor if
    only self references are returned from such functions.

    Any extra arguments passed to the visit method are passed on to child ASTs.
    '''
    def visit(self, ast: AST, *args: Any, **kwargs: Any) -> AST:
        '''
        Dispatch to a visit method in a base class or visit and transform the
        children of the given AST if it is missing.
        '''
        attr = 'visit_' + str(ast.ast_type).replace('ASTType.', '')
        if hasattr(self, attr):
            return getattr(self, attr)(ast, *args, **kwargs)
        return ast.update(**self.visit_children(ast, *args, **kwargs))

    def visit_children(self, ast: AST, *args: Any, **kwargs: Any) -> Dict[str, ASTUpdate]:
        '''
        Visit and transform the children of the given AST.

        Returns
        -------
        The functions returns a dictionary that can be passed to `AST.update`.
        It contains the attributes and values that have been transformed.
        '''
        update: Dict[str, ASTUpdate] = dict()
        for key in ast.child_keys:
            old = getattr(ast, key)
            new = self._dispatch(old, *args, **kwargs)
            if new is not old:
                update[key] = new
        return update

    def visit_sequence(self, sequence: ASTSequence, *args: Any, **kwargs: Any) -> MutableSequence[AST]:
        '''
        Transform a sequence of ASTs returning the same sequnce if there are no
        changes or a list of ASTs otherwise.
        '''
        ret: MutableSequence[AST]
        ret, lst = sequence, []
        for old in sequence:
            lst.append(self(old, *args, **kwargs))
            if lst[-1] is not old:
                ret = lst
        return ret

    def _dispatch(self, ast: Union[None, AST, ASTSequence], *args: Any, **kwargs: Any) -> Union[None, AST, MutableSequence[AST]]:
        '''
        Visit and transform an (optional) AST or a sequence of ASTs.
        '''
        if ast is None:
            return ast

        if isinstance(ast, AST):
            return self.visit(ast, *args, **kwargs) # type: ignore

        if isinstance(ast, abc.Sequence):
            return self.visit_sequence(ast, *args, **kwargs)

        raise TypeError('unexpected type')

    def __call__(self, ast: AST, *args: Any, **kwargs: Any) -> AST:
        '''
        Alternative way to call `Transformer.visit`.
        '''
        return self.visit(ast, *args, **kwargs)

def Id(location: Location, name: str) -> AST:
    '''
    Construct an AST node of type `ASTType.Id`.
    '''
    p_ast = _ffi.new('clingo_ast_t**')
    c_location = _c_location(location)
    _handle_error(_lib.clingo_ast_build(
        _lib.clingo_ast_type_id, p_ast,
        c_location[0],
        _ffi.new('char const[]', name.encode())))
    return AST(p_ast[0])

def Variable(location: Location, name: str) -> AST:
    '''
    Construct an AST node of type `ASTType.Variable`.
    '''
    p_ast = _ffi.new('clingo_ast_t**')
    c_location = _c_location(location)
    _handle_error(_lib.clingo_ast_build(
        _lib.clingo_ast_type_variable, p_ast,
        c_location[0],
        _ffi.new('char const[]', name.encode())))
    return AST(p_ast[0])

def SymbolicTerm(location: Location, symbol: Symbol) -> AST:
    '''
    Construct an AST node of type `ASTType.SymbolicTerm`.
    '''
    p_ast = _ffi.new('clingo_ast_t**')
    c_location = _c_location(location)
    _handle_error(_lib.clingo_ast_build(
        _lib.clingo_ast_type_symbolic_term, p_ast,
        c_location[0],
        _ffi.cast('clingo_symbol_t', symbol._rep)))
    return AST(p_ast[0])

def UnaryOperation(location: Location, operator_type: int, argument: AST) -> AST:
    '''
    Construct an AST node of type `ASTType.UnaryOperation`.
    '''
    p_ast = _ffi.new('clingo_ast_t**')
    c_location = _c_location(location)
    _handle_error(_lib.clingo_ast_build(
        _lib.clingo_ast_type_unary_operation, p_ast,
        c_location[0],
        _ffi.cast('int', operator_type),
        argument._rep))
    return AST(p_ast[0])

def BinaryOperation(location: Location, operator_type: int, left: AST, right: AST) -> AST:
    '''
    Construct an AST node of type `ASTType.BinaryOperation`.
    '''
    p_ast = _ffi.new('clingo_ast_t**')
    c_location = _c_location(location)
    _handle_error(_lib.clingo_ast_build(
        _lib.clingo_ast_type_binary_operation, p_ast,
        c_location[0],
        _ffi.cast('int', operator_type),
        left._rep,
        right._rep))
    return AST(p_ast[0])

def Interval(location: Location, left: AST, right: AST) -> AST:
    '''
    Construct an AST node of type `ASTType.Interval`.
    '''
    p_ast = _ffi.new('clingo_ast_t**')
    c_location = _c_location(location)
    _handle_error(_lib.clingo_ast_build(
        _lib.clingo_ast_type_interval, p_ast,
        c_location[0],
        left._rep,
        right._rep))
    return AST(p_ast[0])

def Function(location: Location, name: str, arguments: Sequence[AST], external: int) -> AST:
    '''
    Construct an AST node of type `ASTType.Function`.
    '''
    p_ast = _ffi.new('clingo_ast_t**')
    c_location = _c_location(location)
    _handle_error(_lib.clingo_ast_build(
        _lib.clingo_ast_type_function, p_ast,
        c_location[0],
        _ffi.new('char const[]', name.encode()),
        _ffi.new('clingo_ast_t*[]', [ x._rep for x in arguments ]),
        _ffi.cast('size_t', len(arguments)),
        _ffi.cast('int', external)))
    return AST(p_ast[0])

def Pool(location: Location, arguments: Sequence[AST]) -> AST:
    '''
    Construct an AST node of type `ASTType.Pool`.
    '''
    p_ast = _ffi.new('clingo_ast_t**')
    c_location = _c_location(location)
    _handle_error(_lib.clingo_ast_build(
        _lib.clingo_ast_type_pool, p_ast,
        c_location[0],
        _ffi.new('clingo_ast_t*[]', [ x._rep for x in arguments ]),
        _ffi.cast('size_t', len(arguments))))
    return AST(p_ast[0])

def CspProduct(location: Location, coefficient: AST, variable: Optional[AST]) -> AST:
    '''
    Construct an AST node of type `ASTType.CspProduct`.
    '''
    p_ast = _ffi.new('clingo_ast_t**')
    c_location = _c_location(location)
    _handle_error(_lib.clingo_ast_build(
        _lib.clingo_ast_type_csp_product, p_ast,
        c_location[0],
        coefficient._rep,
        _ffi.NULL if variable is None else variable._rep))
    return AST(p_ast[0])

def CspSum(location: Location, terms: Sequence[AST]) -> AST:
    '''
    Construct an AST node of type `ASTType.CspSum`.
    '''
    p_ast = _ffi.new('clingo_ast_t**')
    c_location = _c_location(location)
    _handle_error(_lib.clingo_ast_build(
        _lib.clingo_ast_type_csp_sum, p_ast,
        c_location[0],
        _ffi.new('clingo_ast_t*[]', [ x._rep for x in terms ]),
        _ffi.cast('size_t', len(terms))))
    return AST(p_ast[0])

def CspGuard(comparison: int, term: AST) -> AST:
    '''
    Construct an AST node of type `ASTType.CspGuard`.
    '''
    p_ast = _ffi.new('clingo_ast_t**')
    _handle_error(_lib.clingo_ast_build(
        _lib.clingo_ast_type_csp_guard, p_ast,
        _ffi.cast('int', comparison),
        term._rep))
    return AST(p_ast[0])

def BooleanConstant(value: int) -> AST:
    '''
    Construct an AST node of type `ASTType.BooleanConstant`.
    '''
    p_ast = _ffi.new('clingo_ast_t**')
    _handle_error(_lib.clingo_ast_build(
        _lib.clingo_ast_type_boolean_constant, p_ast,
        _ffi.cast('int', value)))
    return AST(p_ast[0])

def SymbolicAtom(symbol: AST) -> AST:
    '''
    Construct an AST node of type `ASTType.SymbolicAtom`.
    '''
    p_ast = _ffi.new('clingo_ast_t**')
    _handle_error(_lib.clingo_ast_build(
        _lib.clingo_ast_type_symbolic_atom, p_ast,
        symbol._rep))
    return AST(p_ast[0])

def Comparison(comparison: int, left: AST, right: AST) -> AST:
    '''
    Construct an AST node of type `ASTType.Comparison`.
    '''
    p_ast = _ffi.new('clingo_ast_t**')
    _handle_error(_lib.clingo_ast_build(
        _lib.clingo_ast_type_comparison, p_ast,
        _ffi.cast('int', comparison),
        left._rep,
        right._rep))
    return AST(p_ast[0])

def CspLiteral(location: Location, term: AST, guards: Sequence[AST]) -> AST:
    '''
    Construct an AST node of type `ASTType.CspLiteral`.
    '''
    p_ast = _ffi.new('clingo_ast_t**')
    c_location = _c_location(location)
    _handle_error(_lib.clingo_ast_build(
        _lib.clingo_ast_type_csp_literal, p_ast,
        c_location[0],
        term._rep,
        _ffi.new('clingo_ast_t*[]', [ x._rep for x in guards ]),
        _ffi.cast('size_t', len(guards))))
    return AST(p_ast[0])

def AggregateGuard(comparison: int, term: AST) -> AST:
    '''
    Construct an AST node of type `ASTType.AggregateGuard`.
    '''
    p_ast = _ffi.new('clingo_ast_t**')
    _handle_error(_lib.clingo_ast_build(
        _lib.clingo_ast_type_aggregate_guard, p_ast,
        _ffi.cast('int', comparison),
        term._rep))
    return AST(p_ast[0])

def ConditionalLiteral(location: Location, literal: AST, condition: Sequence[AST]) -> AST:
    '''
    Construct an AST node of type `ASTType.ConditionalLiteral`.
    '''
    p_ast = _ffi.new('clingo_ast_t**')
    c_location = _c_location(location)
    _handle_error(_lib.clingo_ast_build(
        _lib.clingo_ast_type_conditional_literal, p_ast,
        c_location[0],
        literal._rep,
        _ffi.new('clingo_ast_t*[]', [ x._rep for x in condition ]),
        _ffi.cast('size_t', len(condition))))
    return AST(p_ast[0])

def Aggregate(location: Location, left_guard: Optional[AST], elements: Sequence[AST],
              right_guard: Optional[AST]) -> AST:
    '''
    Construct an AST node of type `ASTType.Aggregate`.
    '''
    p_ast = _ffi.new('clingo_ast_t**')
    c_location = _c_location(location)
    _handle_error(_lib.clingo_ast_build(
        _lib.clingo_ast_type_aggregate, p_ast,
        c_location[0],
        _ffi.NULL if left_guard is None else left_guard._rep,
        _ffi.new('clingo_ast_t*[]', [ x._rep for x in elements ]),
        _ffi.cast('size_t', len(elements)),
        _ffi.NULL if right_guard is None else right_guard._rep))
    return AST(p_ast[0])

def BodyAggregateElement(terms: Sequence[AST], condition: Sequence[AST]) -> AST:
    '''
    Construct an AST node of type `ASTType.BodyAggregateElement`.
    '''
    p_ast = _ffi.new('clingo_ast_t**')
    _handle_error(_lib.clingo_ast_build(
        _lib.clingo_ast_type_body_aggregate_element, p_ast,
        _ffi.new('clingo_ast_t*[]', [ x._rep for x in terms ]),
        _ffi.cast('size_t', len(terms)),
        _ffi.new('clingo_ast_t*[]', [ x._rep for x in condition ]),
        _ffi.cast('size_t', len(condition))))
    return AST(p_ast[0])

def BodyAggregate(location: Location, left_guard: Optional[AST], function: int, elements: Sequence[AST],
                  right_guard: Optional[AST]) -> AST:
    '''
    Construct an AST node of type `ASTType.BodyAggregate`.
    '''
    p_ast = _ffi.new('clingo_ast_t**')
    c_location = _c_location(location)
    _handle_error(_lib.clingo_ast_build(
        _lib.clingo_ast_type_body_aggregate, p_ast,
        c_location[0],
        _ffi.NULL if left_guard is None else left_guard._rep,
        _ffi.cast('int', function),
        _ffi.new('clingo_ast_t*[]', [ x._rep for x in elements ]),
        _ffi.cast('size_t', len(elements)),
        _ffi.NULL if right_guard is None else right_guard._rep))
    return AST(p_ast[0])

def HeadAggregateElement(terms: Sequence[AST], condition: AST) -> AST:
    '''
    Construct an AST node of type `ASTType.HeadAggregateElement`.
    '''
    p_ast = _ffi.new('clingo_ast_t**')
    _handle_error(_lib.clingo_ast_build(
        _lib.clingo_ast_type_head_aggregate_element, p_ast,
        _ffi.new('clingo_ast_t*[]', [ x._rep for x in terms ]),
        _ffi.cast('size_t', len(terms)),
        condition._rep))
    return AST(p_ast[0])

def HeadAggregate(location: Location, left_guard: Optional[AST], function: int, elements: Sequence[AST],
                  right_guard: Optional[AST]) -> AST:
    '''
    Construct an AST node of type `ASTType.HeadAggregate`.
    '''
    p_ast = _ffi.new('clingo_ast_t**')
    c_location = _c_location(location)
    _handle_error(_lib.clingo_ast_build(
        _lib.clingo_ast_type_head_aggregate, p_ast,
        c_location[0],
        _ffi.NULL if left_guard is None else left_guard._rep,
        _ffi.cast('int', function),
        _ffi.new('clingo_ast_t*[]', [ x._rep for x in elements ]),
        _ffi.cast('size_t', len(elements)),
        _ffi.NULL if right_guard is None else right_guard._rep))
    return AST(p_ast[0])

def Disjunction(location: Location, elements: Sequence[AST]) -> AST:
    '''
    Construct an AST node of type `ASTType.Disjunction`.
    '''
    p_ast = _ffi.new('clingo_ast_t**')
    c_location = _c_location(location)
    _handle_error(_lib.clingo_ast_build(
        _lib.clingo_ast_type_disjunction, p_ast,
        c_location[0],
        _ffi.new('clingo_ast_t*[]', [ x._rep for x in elements ]),
        _ffi.cast('size_t', len(elements))))
    return AST(p_ast[0])

def DisjointElement(location: Location, terms: Sequence[AST], term: AST, condition: Sequence[AST]) -> AST:
    '''
    Construct an AST node of type `ASTType.DisjointElement`.
    '''
    p_ast = _ffi.new('clingo_ast_t**')
    c_location = _c_location(location)
    _handle_error(_lib.clingo_ast_build(
        _lib.clingo_ast_type_disjoint_element, p_ast,
        c_location[0],
        _ffi.new('clingo_ast_t*[]', [ x._rep for x in terms ]),
        _ffi.cast('size_t', len(terms)),
        term._rep,
        _ffi.new('clingo_ast_t*[]', [ x._rep for x in condition ]),
        _ffi.cast('size_t', len(condition))))
    return AST(p_ast[0])

def Disjoint(location: Location, elements: Sequence[AST]) -> AST:
    '''
    Construct an AST node of type `ASTType.Disjoint`.
    '''
    p_ast = _ffi.new('clingo_ast_t**')
    c_location = _c_location(location)
    _handle_error(_lib.clingo_ast_build(
        _lib.clingo_ast_type_disjoint, p_ast,
        c_location[0],
        _ffi.new('clingo_ast_t*[]', [ x._rep for x in elements ]),
        _ffi.cast('size_t', len(elements))))
    return AST(p_ast[0])

def TheorySequence(location: Location, sequence_type: int, terms: Sequence[AST]) -> AST:
    '''
    Construct an AST node of type `ASTType.TheorySequence`.
    '''
    p_ast = _ffi.new('clingo_ast_t**')
    c_location = _c_location(location)
    _handle_error(_lib.clingo_ast_build(
        _lib.clingo_ast_type_theory_sequence, p_ast,
        c_location[0],
        _ffi.cast('int', sequence_type),
        _ffi.new('clingo_ast_t*[]', [ x._rep for x in terms ]),
        _ffi.cast('size_t', len(terms))))
    return AST(p_ast[0])

def TheoryFunction(location: Location, name: str, arguments: Sequence[AST]) -> AST:
    '''
    Construct an AST node of type `ASTType.TheoryFunction`.
    '''
    p_ast = _ffi.new('clingo_ast_t**')
    c_location = _c_location(location)
    _handle_error(_lib.clingo_ast_build(
        _lib.clingo_ast_type_theory_function, p_ast,
        c_location[0],
        _ffi.new('char const[]', name.encode()),
        _ffi.new('clingo_ast_t*[]', [ x._rep for x in arguments ]),
        _ffi.cast('size_t', len(arguments))))
    return AST(p_ast[0])

def TheoryUnparsedTermElement(operators: Sequence[str], term: AST) -> AST:
    '''
    Construct an AST node of type `ASTType.TheoryUnparsedTermElement`.
    '''
    p_ast = _ffi.new('clingo_ast_t**')
    c_operators = [ _ffi.new('char[]', x.encode()) for x in operators ]
    _handle_error(_lib.clingo_ast_build(
        _lib.clingo_ast_type_theory_unparsed_term_element, p_ast,
        _ffi.new('char*[]', c_operators),
        _ffi.cast('size_t', len(operators)),
        term._rep))
    return AST(p_ast[0])

def TheoryUnparsedTerm(location: Location, elements: Sequence[AST]) -> AST:
    '''
    Construct an AST node of type `ASTType.TheoryUnparsedTerm`.
    '''
    p_ast = _ffi.new('clingo_ast_t**')
    c_location = _c_location(location)
    _handle_error(_lib.clingo_ast_build(
        _lib.clingo_ast_type_theory_unparsed_term, p_ast,
        c_location[0],
        _ffi.new('clingo_ast_t*[]', [ x._rep for x in elements ]),
        _ffi.cast('size_t', len(elements))))
    return AST(p_ast[0])

def TheoryGuard(operator_name: str, term: AST) -> AST:
    '''
    Construct an AST node of type `ASTType.TheoryGuard`.
    '''
    p_ast = _ffi.new('clingo_ast_t**')
    _handle_error(_lib.clingo_ast_build(
        _lib.clingo_ast_type_theory_guard, p_ast,
        _ffi.new('char const[]', operator_name.encode()),
        term._rep))
    return AST(p_ast[0])

def TheoryAtomElement(terms: Sequence[AST], condition: Sequence[AST]) -> AST:
    '''
    Construct an AST node of type `ASTType.TheoryAtomElement`.
    '''
    p_ast = _ffi.new('clingo_ast_t**')
    _handle_error(_lib.clingo_ast_build(
        _lib.clingo_ast_type_theory_atom_element, p_ast,
        _ffi.new('clingo_ast_t*[]', [ x._rep for x in terms ]),
        _ffi.cast('size_t', len(terms)),
        _ffi.new('clingo_ast_t*[]', [ x._rep for x in condition ]),
        _ffi.cast('size_t', len(condition))))
    return AST(p_ast[0])

def TheoryAtom(location: Location, term: AST, elements: Sequence[AST], guard: Optional[AST]) -> AST:
    '''
    Construct an AST node of type `ASTType.TheoryAtom`.
    '''
    p_ast = _ffi.new('clingo_ast_t**')
    c_location = _c_location(location)
    _handle_error(_lib.clingo_ast_build(
        _lib.clingo_ast_type_theory_atom, p_ast,
        c_location[0],
        term._rep,
        _ffi.new('clingo_ast_t*[]', [ x._rep for x in elements ]),
        _ffi.cast('size_t', len(elements)),
        _ffi.NULL if guard is None else guard._rep))
    return AST(p_ast[0])

def Literal(location: Location, sign: int, atom: AST) -> AST:
    '''
    Construct an AST node of type `ASTType.Literal`.
    '''
    p_ast = _ffi.new('clingo_ast_t**')
    c_location = _c_location(location)
    _handle_error(_lib.clingo_ast_build(
        _lib.clingo_ast_type_literal, p_ast,
        c_location[0],
        _ffi.cast('int', sign),
        atom._rep))
    return AST(p_ast[0])

def TheoryOperatorDefinition(location: Location, name: str, priority: int, operator_type: int) -> AST:
    '''
    Construct an AST node of type `ASTType.TheoryOperatorDefinition`.
    '''
    p_ast = _ffi.new('clingo_ast_t**')
    c_location = _c_location(location)
    _handle_error(_lib.clingo_ast_build(
        _lib.clingo_ast_type_theory_operator_definition, p_ast,
        c_location[0],
        _ffi.new('char const[]', name.encode()),
        _ffi.cast('int', priority),
        _ffi.cast('int', operator_type)))
    return AST(p_ast[0])

def TheoryTermDefinition(location: Location, name: str, operators: Sequence[AST]) -> AST:
    '''
    Construct an AST node of type `ASTType.TheoryTermDefinition`.
    '''
    p_ast = _ffi.new('clingo_ast_t**')
    c_location = _c_location(location)
    _handle_error(_lib.clingo_ast_build(
        _lib.clingo_ast_type_theory_term_definition, p_ast,
        c_location[0],
        _ffi.new('char const[]', name.encode()),
        _ffi.new('clingo_ast_t*[]', [ x._rep for x in operators ]),
        _ffi.cast('size_t', len(operators))))
    return AST(p_ast[0])

def TheoryGuardDefinition(operators: Sequence[str], term: str) -> AST:
    '''
    Construct an AST node of type `ASTType.TheoryGuardDefinition`.
    '''
    p_ast = _ffi.new('clingo_ast_t**')
    c_operators = [ _ffi.new('char[]', x.encode()) for x in operators ]
    _handle_error(_lib.clingo_ast_build(
        _lib.clingo_ast_type_theory_guard_definition, p_ast,
        _ffi.new('char*[]', c_operators),
        _ffi.cast('size_t', len(operators)),
        _ffi.new('char const[]', term.encode())))
    return AST(p_ast[0])

def TheoryAtomDefinition(location: Location, atom_type: int, name: str, arity: int, term: str,
                         guard: Optional[AST]) -> AST:
    '''
    Construct an AST node of type `ASTType.TheoryAtomDefinition`.
    '''
    p_ast = _ffi.new('clingo_ast_t**')
    c_location = _c_location(location)
    _handle_error(_lib.clingo_ast_build(
        _lib.clingo_ast_type_theory_atom_definition, p_ast,
        c_location[0],
        _ffi.cast('int', atom_type),
        _ffi.new('char const[]', name.encode()),
        _ffi.cast('int', arity),
        _ffi.new('char const[]', term.encode()),
        _ffi.NULL if guard is None else guard._rep))
    return AST(p_ast[0])

def Rule(location: Location, head: AST, body: Sequence[AST]) -> AST:
    '''
    Construct an AST node of type `ASTType.Rule`.
    '''
    p_ast = _ffi.new('clingo_ast_t**')
    c_location = _c_location(location)
    _handle_error(_lib.clingo_ast_build(
        _lib.clingo_ast_type_rule, p_ast,
        c_location[0],
        head._rep,
        _ffi.new('clingo_ast_t*[]', [ x._rep for x in body ]),
        _ffi.cast('size_t', len(body))))
    return AST(p_ast[0])

def Definition(location: Location, name: str, value: AST, is_default: int) -> AST:
    '''
    Construct an AST node of type `ASTType.Definition`.
    '''
    p_ast = _ffi.new('clingo_ast_t**')
    c_location = _c_location(location)
    _handle_error(_lib.clingo_ast_build(
        _lib.clingo_ast_type_definition, p_ast,
        c_location[0],
        _ffi.new('char const[]', name.encode()),
        value._rep,
        _ffi.cast('int', is_default)))
    return AST(p_ast[0])

def ShowSignature(location: Location, name: str, arity: int, positive: int, csp: int) -> AST:
    '''
    Construct an AST node of type `ASTType.ShowSignature`.
    '''
    p_ast = _ffi.new('clingo_ast_t**')
    c_location = _c_location(location)
    _handle_error(_lib.clingo_ast_build(
        _lib.clingo_ast_type_show_signature, p_ast,
        c_location[0],
        _ffi.new('char const[]', name.encode()),
        _ffi.cast('int', arity),
        _ffi.cast('int', positive),
        _ffi.cast('int', csp)))
    return AST(p_ast[0])

def ShowTerm(location: Location, term: AST, body: Sequence[AST], csp: int) -> AST:
    '''
    Construct an AST node of type `ASTType.ShowTerm`.
    '''
    p_ast = _ffi.new('clingo_ast_t**')
    c_location = _c_location(location)
    _handle_error(_lib.clingo_ast_build(
        _lib.clingo_ast_type_show_term, p_ast,
        c_location[0],
        term._rep,
        _ffi.new('clingo_ast_t*[]', [ x._rep for x in body ]),
        _ffi.cast('size_t', len(body)),
        _ffi.cast('int', csp)))
    return AST(p_ast[0])

def Minimize(location: Location, weight: AST, priority: AST, terms: Sequence[AST], body: Sequence[AST]) -> AST:
    '''
    Construct an AST node of type `ASTType.Minimize`.
    '''
    p_ast = _ffi.new('clingo_ast_t**')
    c_location = _c_location(location)
    _handle_error(_lib.clingo_ast_build(
        _lib.clingo_ast_type_minimize, p_ast,
        c_location[0],
        weight._rep,
        priority._rep,
        _ffi.new('clingo_ast_t*[]', [ x._rep for x in terms ]),
        _ffi.cast('size_t', len(terms)),
        _ffi.new('clingo_ast_t*[]', [ x._rep for x in body ]),
        _ffi.cast('size_t', len(body))))
    return AST(p_ast[0])

def Script(location: Location, name: str, code: str) -> AST:
    '''
    Construct an AST node of type `ASTType.Script`.
    '''
    p_ast = _ffi.new('clingo_ast_t**')
    c_location = _c_location(location)
    _handle_error(_lib.clingo_ast_build(
        _lib.clingo_ast_type_script, p_ast,
        c_location[0],
        _ffi.new('char const[]', name.encode()),
        _ffi.new('char const[]', code.encode())))
    return AST(p_ast[0])

def Program(location: Location, name: str, parameters: Sequence[AST]) -> AST:
    '''
    Construct an AST node of type `ASTType.Program`.
    '''
    p_ast = _ffi.new('clingo_ast_t**')
    c_location = _c_location(location)
    _handle_error(_lib.clingo_ast_build(
        _lib.clingo_ast_type_program, p_ast,
        c_location[0],
        _ffi.new('char const[]', name.encode()),
        _ffi.new('clingo_ast_t*[]', [ x._rep for x in parameters ]),
        _ffi.cast('size_t', len(parameters))))
    return AST(p_ast[0])

def External(location: Location, atom: AST, body: Sequence[AST], external_type: AST) -> AST:
    '''
    Construct an AST node of type `ASTType.External`.
    '''
    p_ast = _ffi.new('clingo_ast_t**')
    c_location = _c_location(location)
    _handle_error(_lib.clingo_ast_build(
        _lib.clingo_ast_type_external, p_ast,
        c_location[0],
        atom._rep,
        _ffi.new('clingo_ast_t*[]', [ x._rep for x in body ]),
        _ffi.cast('size_t', len(body)),
        external_type._rep))
    return AST(p_ast[0])

def Edge(location: Location, node_u: AST, node_v: AST, body: Sequence[AST]) -> AST:
    '''
    Construct an AST node of type `ASTType.Edge`.
    '''
    p_ast = _ffi.new('clingo_ast_t**')
    c_location = _c_location(location)
    _handle_error(_lib.clingo_ast_build(
        _lib.clingo_ast_type_edge, p_ast,
        c_location[0],
        node_u._rep,
        node_v._rep,
        _ffi.new('clingo_ast_t*[]', [ x._rep for x in body ]),
        _ffi.cast('size_t', len(body))))
    return AST(p_ast[0])

def Heuristic(location: Location, atom: AST, body: Sequence[AST], bias: AST, priority: AST, modifier: AST) -> AST:
    '''
    Construct an AST node of type `ASTType.Heuristic`.
    '''
    p_ast = _ffi.new('clingo_ast_t**')
    c_location = _c_location(location)
    _handle_error(_lib.clingo_ast_build(
        _lib.clingo_ast_type_heuristic, p_ast,
        c_location[0],
        atom._rep,
        _ffi.new('clingo_ast_t*[]', [ x._rep for x in body ]),
        _ffi.cast('size_t', len(body)),
        bias._rep,
        priority._rep,
        modifier._rep))
    return AST(p_ast[0])

def ProjectAtom(location: Location, atom: AST, body: Sequence[AST]) -> AST:
    '''
    Construct an AST node of type `ASTType.ProjectAtom`.
    '''
    p_ast = _ffi.new('clingo_ast_t**')
    c_location = _c_location(location)
    _handle_error(_lib.clingo_ast_build(
        _lib.clingo_ast_type_project_atom, p_ast,
        c_location[0],
        atom._rep,
        _ffi.new('clingo_ast_t*[]', [ x._rep for x in body ]),
        _ffi.cast('size_t', len(body))))
    return AST(p_ast[0])

def ProjectSignature(location: Location, name: str, arity: int, positive: int) -> AST:
    '''
    Construct an AST node of type `ASTType.ProjectSignature`.
    '''
    p_ast = _ffi.new('clingo_ast_t**')
    c_location = _c_location(location)
    _handle_error(_lib.clingo_ast_build(
        _lib.clingo_ast_type_project_signature, p_ast,
        c_location[0],
        _ffi.new('char const[]', name.encode()),
        _ffi.cast('int', arity),
        _ffi.cast('int', positive)))
    return AST(p_ast[0])

def Defined(location: Location, name: str, arity: int, positive: int) -> AST:
    '''
    Construct an AST node of type `ASTType.Defined`.
    '''
    p_ast = _ffi.new('clingo_ast_t**')
    c_location = _c_location(location)
    _handle_error(_lib.clingo_ast_build(
        _lib.clingo_ast_type_defined, p_ast,
        c_location[0],
        _ffi.new('char const[]', name.encode()),
        _ffi.cast('int', arity),
        _ffi.cast('int', positive)))
    return AST(p_ast[0])

def TheoryDefinition(location: Location, name: str, terms: Sequence[AST], atoms: Sequence[AST]) -> AST:
    '''
    Construct an AST node of type `ASTType.TheoryDefinition`.
    '''
    p_ast = _ffi.new('clingo_ast_t**')
    c_location = _c_location(location)
    _handle_error(_lib.clingo_ast_build(
        _lib.clingo_ast_type_theory_definition, p_ast,
        c_location[0],
        _ffi.new('char const[]', name.encode()),
        _ffi.new('clingo_ast_t*[]', [ x._rep for x in terms ]),
        _ffi.cast('size_t', len(terms)),
        _ffi.new('clingo_ast_t*[]', [ x._rep for x in atoms ]),
        _ffi.cast('size_t', len(atoms))))
    return AST(p_ast[0])



#!/usr/bin/python

import re
import os
import os.path
import sys
import subprocess as sp
import difflib as dl
import argparse

parser = argparse.ArgumentParser(description="""
Can be used to run and normalize tests.
Additional options can be passed to clingo by adding them at the end of the command line preceded by '--'.
""")
parser.add_argument('-c', '--clingo', default=None, help="path to clingo executable")

subparsers = parser.add_subparsers(dest="action", help="sub-command --help")
subparsers.required = True
parser_run = subparsers.add_parser('run', help='run all tests')
parser_normalize = subparsers.add_parser('normalize', help='normalize the output of clingo')
parser_normalize.add_argument("file")

argv = sys.argv[1:]
extra_argv = []
if "--" in argv:
    extra_argv = argv[argv.index("--")+1:]
    argv       = argv[:argv.index("--")]

parse_ret = parser.parse_args(argv)

if parse_ret.action is None:
    print (parser.usage)
    exit(0)

clingo = parse_ret.clingo

def find_clingo():
    clingos = [
        "build/debug/bin/clingo",
        "build/release/bin/clingo",
        "build/bin/clingo",
        "build/cmake/bin/clingo",
        "build/bin/Debug/clingo.exe",
        "build/bin/Release/clingo.exe",
        ]
    for x in clingos:
        x = os.path.normpath("{}/../../../{}".format(wd, x))
        if os.path.exists(x):
            return x
    return None

wd = os.path.normpath(os.path.dirname(__file__))
if clingo is None:
    clingo = find_clingo()

if clingo is None:
    print ("no usable clingo version found")
    exit(1)

def normalize(out):
    state=0
    current=[]
    step=0
    result="ERROR"
    norm=[]
    for line in out.split("\n"):
        if state == 1:
            if step > 0:
                norm.append("Step: {}".format(step))
                models = []
                for model in current:
                    models.append(" ".join(sorted(model)))
                for model in sorted(models):
                    norm.append(model)
            step += 1
            state = 0
            current = []
        if state == 2:
            current.append(line.strip().split(" "))
            state=0
        elif line.startswith("Solving..."):
            state = 1
        elif line.startswith("Answer: "):
            state = 2
        elif line.startswith("SATISFIABLE"):
            result="SAT"
            state=1
        elif line.startswith("UNSATISFIABLE"):
            result="UNSAT"
            state=1
        elif line.startswith("UNKNOWN"):
            result="UNKNOWN"
            state=1
        elif line.startswith("OPTIMUM FOUND"):
            result="OPTIMUM FOUND"
            state=1
    norm.append(result)
    norm.append("")
    return "\n".join(norm)

if parse_ret.action == "normalize":
    args = [clingo, "0", parse_ret.file, "-Wnone"]
    b = os.path.splitext(parse_ret.file)[0]
    if os.path.exists(b + ".cmd"):
        with open(b + ".cmd", 'rU') as cmd_file:
            for x in cmd_file:
                args.extend(x.strip().split())
    args.extend(extra_argv)
    out, err = sp.Popen(args, stderr=sp.PIPE, stdout=sp.PIPE, universal_newlines=True).communicate()
    sys.stdout.write(normalize(out))
    exit(0)
if parse_ret.action == "run":
    total  = 0
    failed = 0

    out, err = sp.Popen([clingo, "--version"], stderr=sp.PIPE, stdout=sp.PIPE, universal_newlines=True).communicate()
    with_python  = out.find("with Python") > 0
    with_lua     = out.find("with Lua") > 0
    with_threads = out.find("WITH_THREADS=1") > 0
    for root, dirs, files in os.walk(wd):
        for f in sorted(files):
            if f.endswith(".lp"):
                b = os.path.join(root, f[:-3])
                with open(b + ".lp", 'rU') as inst_file:
                    inst = inst_file.read()
                    if (not with_python and re.search(r"#script[ ]*\(python\)", inst)) or \
                       (not with_lua and re.search(r"#script[ ]*\(lua\)", inst)) or \
                       (not with_threads and re.search("async_=", inst)) or \
                       (not with_threads and re.search("solve_async", inst)):
                        continue

                total+= 1
                sys.stdout.flush()

                args = [clingo, "0", b + ".lp", "-Wnone"]
                if os.path.exists(b + ".cmd"):
                    with open(b + ".cmd", 'rU') as cmd_file:
                        for x in cmd_file:
                            args.extend(x.strip().split())
                args.extend(extra_argv)
                out, err = sp.Popen(args, stderr=sp.PIPE, stdout=sp.PIPE, universal_newlines=True).communicate()
                norm = normalize(out)
                with open(b + ".sol", 'rU') as sol_file:
                    sol  = sol_file.read()
                if norm != sol:
                    failed+= 1
                    print
                    print ("-" * 79)
                    print (" ".join(args))
                    print ("." * 79)
                    print
                    print ("FAILED:")
                    d = dl.Differ()
                    for line in list(d.compare(sol.splitlines(), norm.splitlines())):
                        if not line.startswith(" "):
                            print (line)
                    print
                    print ("." * 79)
                    print
                    print ("STDOUT:")
                    print
                    print (out)
                    print
                    print ("." * 79)
                    print
                    print ("STDERR:")
                    print
                    print (err)
                    print
                sys.stdout.flush()

    print ("=" * 79)
    if failed > 0:
        print ("Some tests failed ({} of {} test cases)".format(failed, total))
        print
        exit(1)
    else:
        print ("All tests passed ({} test cases)".format(total))
        print
        exit(0)




'''
This module contains functions to add custom scripts, which can be embedded
into logic programs.

Examples
--------
The following example shows how to add a script that works the same way as
clingo's embedded Python script:
```python
>>> from clingo.script import Script, register_script
>>> from clingo.control import Control
>>>
>>> import __main__
>>>
>>> class MyPythonScript(Script):
...     def execute(self, location, code):
...         exec(code, __main__.__dict__, __main__.__dict__)
...     def call(self, location, name, arguments):
...         return getattr(__main__, name)(*arguments)
...     def callable(self, name):
...         return name in __main__.__dict__ and callable(__main__.__dict__[name])
...
>>> register_script('mypython', MyPythonScript())
>>>
>>> ctl = Control()
>>> ctl.add('base', [], """
... #script(mypython)
... from clingo.symbol import Number
... def func(a):
...     return Number(a.number + 1)
... #end.
... a(@func(1)).
... """)
>>>
>>> ctl.ground([('base',[])])
>>> ctl.solve(on_model=print)
a(2)
```
'''

from platform import python_version
from abc import ABCMeta, abstractmethod
from typing import Any, List, Iterable, Tuple, Union
from collections.abc import Iterable as IterableABC
from traceback import format_exception
from clingo._internal import _c_call, _ffi, _handle_error, _lib
from clingo.control import Control
from clingo.symbol import Symbol
from clingo.ast import Location, _py_location

try:
    import __main__ # type: ignore
except ImportError:
    # Note: pypy does not create a main module if embedded
    import sys
    import types
    sys.modules['__main__'] = types.ModuleType('__main__', 'the main module')
    import __main__ # type: ignore

__all__ = [ 'Script', 'enable_python', 'register_script' ]

def _cb_error_top_level(exception, exc_value, traceback):
    msg = "".join(format_exception(exception, exc_value, traceback))
    _lib.clingo_set_error(_lib.clingo_error_runtime, msg.encode())
    return False


class Script(metaclass=ABCMeta):
    '''
    This interface can be implemented to embed custom scripting languages into
    logic programs.
    '''
    @abstractmethod
    def execute(self, location: Location, code: str) -> None:
        '''
        Execute the given source code.

        Parameters
        ----------
        location
            The location of the code.
        code
            The code to execute.
        '''

    @abstractmethod
    def call(self, location: Location, name: str, arguments: Iterable[Symbol]) -> Union[Iterable[Symbol], Symbol]:
        '''
        Call the function with the given name and arguments.

        Parameters
        ----------
        location
            From where in the logic program the function was called.
        name
            The name of the function.
        arguments
            The arguments to the function.

        Returns
        -------
        The resulting pool of symbols.
        '''

    @abstractmethod
    def callable(self, name: str) -> bool:
        '''
        Check there is a function with the given name.

        Parameters
        ----------
        name
            The name of the function.

        Returns
        -------
        Whether the function is callable.
        '''

    def main(self, control: Control) -> None:
        '''
        Run the main function.

        This function exisits primarily for internal purposes and does not need
        to be implemented.

        Parameters
        ----------
        control
            Control object to pass to the main function.
        '''

class _PythonScript(Script):
    def execute(self, location, code):
        exec(code, __main__.__dict__, __main__.__dict__) # pylint: disable=exec-used

    def call(self, location, name, arguments):
        fun = getattr(__main__, name)
        return fun(*arguments)

    def callable(self, name):
        return name in __main__.__dict__ and callable(__main__.__dict__[name])

    def main(self, control):
        __main__.main(control) # pylint: disable=c-extension-no-member

_PYTHON_SCRIPT = _PythonScript()
_GLOBAL_SCRIPTS: List[Tuple[Script, Any]] = []


@_ffi.def_extern(onerror=_cb_error_top_level, name='pyclingo_script_execute')
def _pyclingo_script_execute(location, code, data):
    script: Script = _ffi.from_handle(data)
    script.execute(_py_location(location), _ffi.string(code).decode())
    return True

@_ffi.def_extern(onerror=_cb_error_top_level, name='pyclingo_script_call')
def _pyclingo_script_call(location, name, arguments, size, symbol_callback, symbol_callback_data, data):
    script: Script = _ffi.from_handle(data)
    symbol_callback = _ffi.cast('clingo_symbol_callback_t', symbol_callback)
    arguments = _ffi.cast('clingo_symbol_t*', arguments)
    py_name = _ffi.string(name).decode()
    py_args = [Symbol(arguments[i]) for i in range(size)]

    ret = script.call(_py_location(location), py_name, py_args)
    symbols = list(ret) if isinstance(ret, IterableABC) else [ret]

    c_symbols = _ffi.new('clingo_symbol_t[]', len(symbols))
    for i, sym in enumerate(symbols):
        c_symbols[i] = sym._rep # pylint: disable=protected-access
    _handle_error(symbol_callback(c_symbols, len(symbols), symbol_callback_data))
    return True

@_ffi.def_extern(onerror=_cb_error_top_level, name='pyclingo_script_callable')
def _pyclingo_script_callable(name, ret, data):
    script: Script = _ffi.from_handle(data)
    py_name = _ffi.string(name).decode()
    ret[0] = script.callable(py_name)
    return True

@_ffi.def_extern(onerror=_cb_error_top_level, name='pyclingo_script_main')
def _pyclingo_script_main(ctl, data):
    script: Script = _ffi.from_handle(data)
    script.main(Control(_ffi.cast('clingo_control_t*', ctl)))
    return True


@_ffi.def_extern(onerror=_cb_error_top_level, name='pyclingo_execute')
def _pyclingo_execute(location, code, data):
    assert data == _ffi.NULL
    return _pyclingo_script_execute(_ffi.cast('clingo_location_t*', location),
                                   code,
                                   _ffi.new_handle(_PYTHON_SCRIPT))

@_ffi.def_extern(onerror=_cb_error_top_level, name='pyclingo_call')
def _pyclingo_call(location, name, arguments, size, symbol_callback, symbol_callback_data, data):
    assert data == _ffi.NULL
    return _pyclingo_script_call(_ffi.cast('clingo_location_t*', location),
                                 name,
                                 arguments, size,
                                 symbol_callback, symbol_callback_data,
                                 _ffi.new_handle(_PYTHON_SCRIPT))

@_ffi.def_extern(onerror=_cb_error_top_level, name='pyclingo_callable')
def _pyclingo_callable(name, ret, data):
    assert data == _ffi.NULL
    return _pyclingo_script_callable(name,
                                     ret,
                                     _ffi.new_handle(_PYTHON_SCRIPT))

@_ffi.def_extern(onerror=_cb_error_top_level, name='pyclingo_main')
def _pyclingo_main(ctl, data):
    assert data == _ffi.NULL
    return _pyclingo_script_main(_ffi.cast('clingo_control_t*', ctl),
                                 _ffi.new_handle(_PYTHON_SCRIPT))


def register_script(name: str, script: Script, version: str = '1.0.0') -> None:
    '''
    Registers a script language which can then be embedded into a logic
    program.

    Parameters
    ----------
    name
        The name of the script. This name can then be used in the script
        statement in a logic program.
    script
        The class to register.
    version
        The version of the script.
    '''
    c_version = _c_call('char const*', _lib.clingo_add_string, version.encode())
    c_script = _ffi.new('clingo_script_t*')
    c_script[0].execute = _ffi.cast('void*', _lib.pyclingo_script_execute)
    c_script[0].call = _ffi.cast('void*', _lib.pyclingo_script_call)
    c_script[0].callable = _ffi.cast('void*', _lib.pyclingo_script_callable)
    c_script[0].main = _ffi.cast('void*', _lib.pyclingo_script_main)
    c_script[0].free = _ffi.NULL
    c_script[0].version = c_version
    data = _ffi.new_handle(script)
    _GLOBAL_SCRIPTS.append((script, data))
    _handle_error(_lib.clingo_register_script(name.encode(), c_script, data))

def enable_python() -> None:
    '''
    This function can be called to enable evaluation of Python scripts in logic
    programs.

    By default evaluation is only enabled in the clingo executable but not in
    the Python module.
    '''
    c_version = _c_call('char const*', _lib.clingo_add_string, python_version().encode())
    c_script = _ffi.new('clingo_script_t*')
    c_script[0].execute = _ffi.cast('void*', _lib.pyclingo_execute)
    c_script[0].call = _ffi.cast('void*', _lib.pyclingo_call)
    c_script[0].callable = _ffi.cast('void*', _lib.pyclingo_callable)
    c_script[0].main = _ffi.cast('void*', _lib.pyclingo_main)
    c_script[0].free = _ffi.NULL
    c_script[0].version = c_version
    _handle_error(_lib.clingo_register_script("python".encode(), c_script, _ffi.NULL))



'''
Core functionality used throught the clingo package.
'''

from typing import Callable, Tuple
from enum import Enum

from ._internal import _cb_error_panic, _ffi, _lib, _to_str

__all__ = [ 'Logger', 'MessageCode', 'TruthValue', 'version' ]

def version() -> Tuple[int, int, int]:
    '''
    Clingo's version as a tuple `(major, minor, revision)`.
    '''
    p_major = _ffi.new('int*')
    p_minor = _ffi.new('int*')
    p_revision = _ffi.new('int*')
    _lib.clingo_version(p_major, p_minor, p_revision)
    return p_major[0], p_minor[0], p_revision[0]

class MessageCode(Enum):
    '''
    Enumeration of messages codes.
    '''
    AtomUndefined = _lib.clingo_warning_atom_undefined
    '''
    Informs about an undefined atom in program.
    '''
    FileIncluded = _lib.clingo_warning_file_included
    '''
    Indicates that the same file was included multiple times.
    '''
    GlobalVariable = _lib.clingo_warning_global_variable
    '''
    Informs about a global variable in a tuple of an aggregate element.
    '''
    OperationUndefined = _lib.clingo_warning_operation_undefined
    '''
    Inform about an undefined arithmetic operation or unsupported weight of an
    aggregate.
    '''
    Other = _lib.clingo_warning_other
    '''
    Reports other kinds of messages.
    '''
    RuntimeError = _lib.clingo_warning_runtime_error
    '''
    To report multiple errors; a corresponding runtime error is raised later.
    '''
    VariableUnbounded = _lib.clingo_warning_variable_unbounded
    '''
    Informs about a CSP variable with an unbounded domain.
    '''

Logger = Callable[[MessageCode, str], None]

@_ffi.def_extern(onerror=_cb_error_panic, name='pyclingo_logger_callback')
def _pyclingo_logger_callback(code, message, data):
    '''
    Low-level logger callback.
    '''
    handler = _ffi.from_handle(data)
    handler(MessageCode(code), _to_str(message))

class TruthValue(Enum):
    '''
    Enumeration of the different truth values.
    '''
    False_ = _lib.clingo_external_type_false
    '''
    Represents truth value true.
    '''
    Free = _lib.clingo_external_type_free
    '''
    Represents absence of a truth value.
    '''
    True_ = _lib.clingo_external_type_true
    '''
    Represents truth value true.
    '''
    Release = _lib.clingo_external_type_release
    '''
    Indicates that an atom is to be released.
    '''



'''
Module providing functions and classes to control the grounding and solving
process.

Next follow some concepts used throughout the modules of the clingo package.

Terms
-----
Terms without variables and interpreted functions are called symbols in the
following. They are wrapped in the `clingo.symbol.Symbol` class.

Symbolic Atoms and Literals
---------------------------
*Symbolic atoms* without variables and interpreted functions, which appear in
ground logic programs, are captured using the `clingo.symbol.Symbol` class.
They must be of type `clingo.symbol.SymbolType.Function`. Furthermore, some
functions accept *symbolic literals*, which are represented as pairs of symbols
and Booleans. The Boolean stands for the sign of the literal (`True` for
positive and `False` for negative).

Program Literals
----------------
When passing a ground logic program to a solver, clingo does not use a human
readable textual representation but the aspif format. The literals in this
format are called *program literals*. They are non-zero integers associated
with symbolic atoms, theory atoms, and also without any association if they are
used to translate complex language constructs not directly representable in
aspif format. The sign of a program literal is used to represent default
negation. Symbolic atoms can be mapped to program literals using the
`clingo.symbolic_atoms` module and theory atoms using the `clingo.theory_atoms`
module. Note that symbolic and theory atoms can share the same program
literals. Finally, the `clingo.backend` module can also be used to introduce
fresh symbolic atoms and program literals.

Solver Literals
---------------
Before solving, programs in aspif format are translated to an internal solver
representation, where program literals are again mapped to non-zero integers,
so called *solver literals*. The `clingo.propagator.PropagateInit.solver_literal`
function can be used to map program literals to solver literals. Note that
different program literals can share the same solver literal.

Embedded Python Code
--------------------
If the clingo application is build with Python support, clingo will also be
able to execute Python code embedded in logic programs. Functions defined in a
Python script block are callable during the instantiation process using
`@`-syntax. The default grounding/solving process can be customized if a main
function is provided.

## Examples

The first example shows how to use the clingo module from Python.

    >>> from clingo.symbol import Number
    >>> from clingo.control import Control
    >>>
    >>> class Context:
    ...     def inc(self, x):
    ...         return Number(x.number + 1)
    ...     def seq(self, x, y):
    ...         return [x, y]
    ...
    >>> def on_model(m):
    ...     print (m)
    ...
    >>> ctl = Control()
    >>> ctl.add("base", [], """\\
    ... p(@inc(10)).
    ... q(@seq(1,2)).
    ... """)
    >>> ctl.ground([("base", [])], context=Context())
    >>> print(ctl.solve(on_model=on_model))
    p(11) q(1) q(2)
    SAT

The second example shows how to use Python code from clingo.

    #script (python)

    from clingo.symbol import Number

    class Context:
        def inc(self, x):
            return Number(x.number)

        def seq(self, x, y):
            return [x, y]

    def main(prg):
        prg.ground([("base", [])], context=Context())
        prg.solve()

    #end.

    p(@inc(10)).
    q(@seq(1,2)).
'''

from .core import *
from .symbol import *
from .symbolic_atoms import *
from .theory_atoms import *
from .solving import *
from .propagator import *
from .backend import *
from .configuration import *
from .statistics import *
from .control import *
from .application import *

__version__ = '.'.join(map(str, version()))



from distutils.sysconfig import get_python_lib, get_config_vars
import sys
if sys.argv[1] == "prefix":
    print(get_python_lib(True, False, sys.argv[2] if len(sys.argv) > 2 else None))
elif sys.argv[1] == "suffix":
    SO, SOABI, EXT_SUFFIX = get_config_vars("SO", "SOABI", "EXT_SUFFIX")
    if EXT_SUFFIX is not None:
        ext = EXT_SUFFIX
    elif SOABI is not None:
        ext = ''.join('.', SOABI, SO)
    else:
        ext = SO
    print(ext)



'''
Test clingo's Symbol class.
'''
from unittest import TestCase

from clingo import Function, Infimum, Number, String, Supremum, SymbolType, Tuple_, parse_term

class TestSymbol(TestCase):
    '''
    Tests for the symbol class.
    '''

    def test_parse(self):
        '''
        Test term parsing.
        '''
        self.assertEqual(str(parse_term("p(3)")), "p(3)")
        log = []
        self.assertRaises(RuntimeError, parse_term, "p(1/0)", lambda code, message: log.append((code, message)))
        self.assertEqual(len(log), 0)


    def test_str(self):
        '''
        Test string representation of symbols.
        '''
        self.assertEqual(str(Number(10)), "10")

    def test_repr(self):
        '''
        Test representation of symbols.
        '''
        self.assertEqual(
            repr(Function('test', [Number(10), Infimum, Supremum, String("test")], False)),
            "Function('test', [Number(10), Infimum, Supremum, String('test')], False)")

    def test_cmp(self):
        '''
        Test hashing and comparisons.
        '''
        self.assertEqual(hash(Number(10)), hash(Number(10)))
        self.assertNotEqual(hash(Number(1)), hash(Number(2)))
        self.assertTrue(Number(10) == Number(10))
        self.assertTrue(Number(1) != Number(2))
        self.assertTrue(Number(1) < Number(2))
        self.assertFalse(Number(2) < Number(1))
        self.assertTrue(Number(2) > Number(1))
        self.assertFalse(Number(1) > Number(2))

    def test_match(self):
        '''
        Test symbol matching.
        '''
        self.assertTrue(Function('f', [Number(1)]).match("f", 1))
        self.assertFalse(Function('f', [Number(1), Number(2)]).match("f", 1))
        self.assertFalse(Number(1).match("f", 1))

    def test_number(self):
        '''
        Test numbers.
        '''
        n = Number(1)
        self.assertEqual(n.number, 1)
        self.assertEqual(n.type, SymbolType.Number)

    def test_function(self):
        '''
        Test functions.
        '''
        f = Function("f", [Number(1)], False)
        self.assertEqual(f.arguments, [Number(1)])
        self.assertFalse(f.positive)
        self.assertTrue(f.negative)
        self.assertEqual(f.name, "f")
        self.assertEqual(f.type, SymbolType.Function)
        self.assertEqual(Tuple_([]).type, SymbolType.Function)

    def test_infsup(self):
        '''
        Test infimum and supremum.
        '''
        self.assertEqual(Infimum.type, SymbolType.Infimum)
        self.assertEqual(Supremum.type, SymbolType.Supremum)

    def test_string(self):
        '''
        Test strings.
        '''
        s = String("blub")
        self.assertEqual(s.string, "blub")
        self.assertEqual(s.type, SymbolType.String)



'''
Helpers to simplify testing.
'''

from unittest import TestCase

from .. import SolveResult, parse_term

def _p(*models):
    return [[parse_term(symbol) for symbol in model] for model in models]

class _MCB:
    # pylint: disable=missing-function-docstring
    def __init__(self):
        self._models = []
        self._core = None
        self.last = None

    def on_core(self, c):
        self._core = c

    def on_model(self, m):
        self.last = (m.type, sorted(m.symbols(shown=True)))
        self._models.append(self.last[1])

    @property
    def core(self):
        return sorted(self._core)

    @property
    def models(self):
        return sorted(self._models)

def _check_sat(case: TestCase, ret: SolveResult) -> None:
    case.assertTrue(ret.satisfiable is True)
    case.assertTrue(ret.unsatisfiable is False)
    case.assertTrue(ret.unknown is False)
    case.assertTrue(ret.exhausted is True)



'''
Internal functions and classes of the clingo module.
'''

from os import _exit
from traceback import print_exception
import os
import sys
try:
    _FLAGS = None
    # In the pip module the library also exports the clingo symbols, which
    # should be globally available for other libraries depending on clingo.
    if hasattr(sys, 'setdlopenflags'):
        _FLAGS = sys.getdlopenflags()
        sys.setdlopenflags(os.RTLD_LAZY|os.RTLD_GLOBAL)
    try:
        # Note: imported first to correctly handle the embedded case
        from _clingo import ffi as _ffi, lib as _lib # type: ignore # pylint: disable=no-name-in-module
    except ImportError:
        from ._clingo import ffi as _ffi, lib as _lib # type: ignore # pylint: disable=no-name-in-module
finally:
    if _FLAGS is not None:
        sys.setdlopenflags(_FLAGS)

def _str(f_size, f_str, *args, handler=None):
    p_size = _ffi.new('size_t*')
    _handle_error(f_size(*args, p_size), handler)
    p_str = _ffi.new('char[]', p_size[0])
    _handle_error(f_str(*args, p_str, p_size[0]), handler)
    return _ffi.string(p_str).decode()

def _c_call(c_type, c_fun, *args, handler=None):
    '''
    Helper to simplify calling C functions where the last parameter is a
    reference to the return value.
    '''
    if isinstance(c_type, str):
        p_ret = _ffi.new(f'{c_type}*')
    else:
        p_ret = c_type
    _handle_error(c_fun(*args, p_ret), handler)
    return p_ret[0]

def _c_call2(c_type1, c_type2, c_fun, *args, handler=None):
    '''
    Helper to simplify calling C functions where the last two parameters are a
    reference to the return value.
    '''
    p_ret1 = _ffi.new(f'{c_type1}*')
    p_ret2 = _ffi.new(f'{c_type2}*')
    _handle_error(c_fun(*args, p_ret1, p_ret2), handler)
    return p_ret1[0], p_ret2[0]

def _to_str(c_str) -> str:
    return _ffi.string(c_str).decode()

def _handle_error(ret: bool, handler=None):
    if not ret:
        code = _lib.clingo_error_code()
        if code == _lib.clingo_error_unknown and handler is not None and handler.error is not None:
            raise handler.error[0](handler.error[1]).with_traceback(handler.error[2])
        msg = _ffi.string(_lib.clingo_error_message()).decode()
        if code == _lib.clingo_error_bad_alloc:
            raise MemoryError(msg)
        raise RuntimeError(msg)

def _cb_error_handler(param: str):
    def handler(exception, exc_value, traceback) -> bool:
        if traceback is not None:
            handler = _ffi.from_handle(traceback.tb_frame.f_locals[param])
            handler.error = (exception, exc_value, traceback)
            _lib.clingo_set_error(_lib.clingo_error_unknown, str(exc_value).encode())
        else:
            _lib.clingo_set_error(_lib.clingo_error_runtime, "error in callback".encode())
        return False
    return handler

def _cb_error_panic(exception, exc_value, traceback):
    print_exception(exception, exc_value, traceback)
    sys.stderr.write('PANIC: exception in nothrow scope')
    _exit(1)

def _cb_error_print(exception, exc_value, traceback):
    print_exception(exception, exc_value, traceback)
    return False

class _Error:
    '''
    Class to store an error in a unique location.
    '''
    def __init__(self):
        self._error = None

    def clear(self):
        '''
        Clears the last error set.
        '''
        self._error = None

    @property
    def error(self):
        '''
        Return the last error set.
        '''
        # pylint: disable=protected-access,missing-function-docstring
        return self._error

    @error.setter
    def error(self, value):
        '''
        Set an error if no error has been set before.

        This function is thread-safe.
        '''
        # pylint: disable=protected-access
        self._error = value

class _CBData:
    '''
    The class stores the data object that should be passed to a callback as
    well as provides the means to set an error while a callback is running.
    '''
    def __init__(self, data, error):
        self.data = data
        self._error = error

    @property
    def error(self):
        '''
        Get the last error in the underlying error object.
        '''
        return self._error.error

    @error.setter
    def error(self, value):
        '''
        Set error in the underlying error object.
        '''
        self._error.error = value

def _overwritten(base, obj, function):
    return hasattr(obj, function) and (
        not hasattr(base, function) or
        getattr(base, function) is not getattr(obj.__class__, function, None))



'''
Functions and classes for symbol manipulation.

Examples
--------

    >>> from clingo.symbol import Function, Number, parse_term
    >>>
    >>> num = Number(42)
    >>> num.number
    42
    >>> fun = Function("f", [num])
    >>> fun.name
    'f'
    >>> [ str(arg) for arg in fun.arguments ]
    ['42']
    >>> parse_term(str(fun)) == fun
    True
    >>> clingo.parse_term('p(1+2)')
    p(3)
'''

from typing import Callable, List, Optional, Sequence

from enum import Enum
from functools import total_ordering

from ._internal import _c_call, _c_call2, _ffi, _lib, _str, _to_str
from .core import MessageCode

__all__ = [ 'Function', 'Infimum', 'Number', 'String', 'Supremum', 'Symbol',
            'SymbolType', 'Tuple_', 'parse_term']

class SymbolType(Enum):
    '''
    Enumeration of symbols types.
    '''
    Function = _lib.clingo_symbol_type_function
    '''
    A function symbol, e.g., `c`, `(1,"a")`, or `f(1,"a")`.
    '''
    Infimum  = _lib.clingo_symbol_type_infimum
    '''
    The `#inf` symbol.
    '''
    Number  = _lib.clingo_symbol_type_number
    '''
    A numeric symbol, e.g., `1`.
    '''
    String  = _lib.clingo_symbol_type_string
    '''
    A string symbol, e.g., `"a"`.
    '''
    Supremum  = _lib.clingo_symbol_type_supremum
    '''
    The `#sup` symbol
    '''

@total_ordering
class Symbol:
    '''
    Represents a gringo symbol.

    This includes numbers, strings, functions (including constants with
    `len(arguments) == 0` and tuples with `len(name) == 0`), `#inf` and `#sup`.

    Symbol objects implement Python's rich comparison operators and are ordered
    like in gringo. They can also be used as keys in dictionaries. Their string
    representation corresponds to their gringo representation.

    Notes
    -----
    Note that this class does not have a constructor. Instead there are the
    functions `Number`, `String`, `Tuple_`, and `Function` to construct symbol
    objects or the preconstructed symbols `Infimum` and `Supremum`.
    '''
    __slots__ = ('_rep',)

    def __init__(self, rep):
        self._rep = rep

    def __str__(self) -> str:
        return _str(_lib.clingo_symbol_to_string_size, _lib.clingo_symbol_to_string, self._rep)

    def __repr__(self) -> str:
        if self.type == SymbolType.Infimum:
            return 'Infimum'
        if self.type == SymbolType.Supremum:
            return 'Supremum'
        if self.type == SymbolType.Number:
            return f'Number({self.number!r})'
        if self.type == SymbolType.String:
            return f'String({self.string!r})'
        assert self.type == SymbolType.Function
        return f'Function({self.name!r}, {self.arguments!r}, {self.positive!r})'

    def __hash__(self) -> int:
        return _lib.clingo_symbol_hash(self._rep)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Symbol):
            return NotImplemented
        return _lib.clingo_symbol_is_equal_to(self._rep, other._rep)

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, Symbol):
            return NotImplemented
        return _lib.clingo_symbol_is_less_than(self._rep, other._rep)

    def match(self, name: str, arity: int, positive: bool = True) -> bool:
        '''
        Check if this is a function symbol with the given signature.

        Parameters
        ----------
        name
            The name of the function.

        arity
            The arity of the function.

        positive
            Whether to match positive or negative signatures.

        Returns
        -------
        Whether the function matches.
        '''
        return (self.type == SymbolType.Function and
                self.positive == positive and
                self.name == name and
                len(self.arguments) == arity)

    @property
    def arguments(self) -> List['Symbol']:
        '''
        The arguments of a function.
        '''
        args, size = _c_call2('clingo_symbol_t*', 'size_t', _lib.clingo_symbol_arguments, self._rep)
        return [Symbol(args[i]) for i in range(size)]

    @property
    def name(self) -> str:
        '''
        The name of a function.
        '''
        return _to_str(_c_call('char*', _lib.clingo_symbol_name, self._rep))

    @property
    def negative(self) -> bool:
        '''
        The inverted sign of a function.
        '''
        return _c_call('bool', _lib.clingo_symbol_is_negative, self._rep)

    @property
    def number(self) -> int:
        '''
        The value of a number.
        '''
        return _c_call('int', _lib.clingo_symbol_number, self._rep)

    @property
    def positive(self) -> bool:
        '''
        The sign of a function.
        '''
        return _c_call('bool', _lib.clingo_symbol_is_positive, self._rep)

    @property
    def string(self) -> str:
        '''
        The value of a string.
        '''
        return _to_str(_c_call('char*', _lib.clingo_symbol_string, self._rep))

    @property
    def type(self) -> SymbolType:
        '''
        The type of the symbol.
        '''
        return SymbolType(_lib.clingo_symbol_type(self._rep))

def Function(name: str, arguments: Sequence[Symbol]=[], positive: bool=True) -> Symbol:
    '''
    Construct a function symbol.

    This includes constants and tuples. Constants have an empty argument list
    and tuples have an empty name. Functions can represent classically negated
    atoms. Argument `positive` has to be set to false to represent such atoms.

    Parameters
    ----------
    name
        The name of the function (empty for tuples).
    arguments
        The arguments in form of a list of symbols.
    positive
        The sign of the function (tuples must not have signs).
    '''
    # pylint: disable=protected-access,invalid-name,dangerous-default-value
    c_args = _ffi.new('clingo_symbol_t[]', len(arguments))
    for i, arg in enumerate(arguments):
        c_args[i] = arg._rep
    sym = _c_call('clingo_symbol_t', _lib.clingo_symbol_create_function,
                  name.encode(), c_args, len(arguments), positive)
    return Symbol(sym)

def Number(number: int) -> Symbol:
    '''
    Construct a numeric symbol given a number.

    Parameters
    ----------
    number
        The given number.
    '''
    # pylint: disable=invalid-name
    p_rep = _ffi.new('clingo_symbol_t*')
    _lib.clingo_symbol_create_number(number, p_rep)
    return Symbol(p_rep[0])

def String(string: str) -> Symbol:
    '''
    Construct a string symbol given a string.

    Parameters
    ----------
    string
        The given string.
    '''
    # pylint: disable=invalid-name
    return Symbol(_c_call('clingo_symbol_t', _lib.clingo_symbol_create_string, string.encode()))

def Tuple_(arguments: Sequence[Symbol]) -> Symbol:
    '''
    A shortcut for `Function("", arguments)`.

    Parameters
    ----------
    arguments
        The arguments in form of a list of symbols.

    See Also
    --------
    Function
    '''
    # pylint: disable=invalid-name
    return Function("", arguments)

_p_infimum = _ffi.new('clingo_symbol_t*')
_p_supremum = _ffi.new('clingo_symbol_t*')
_lib.clingo_symbol_create_infimum(_p_infimum)
_lib.clingo_symbol_create_supremum(_p_supremum)

Infimum: Symbol = Symbol(_p_infimum[0])
Supremum: Symbol = Symbol(_p_supremum[0])

def parse_term(string: str, logger: Optional[Callable[[MessageCode,str],None]]=None, message_limit: int=20) -> Symbol:
    '''
    Parse the given string using gringo's term parser for ground terms.

    The function also evaluates arithmetic functions.

    Parameters
    ----------
    string
        The string to be parsed.
    logger
        Function to intercept messages normally printed to standard error.
    message_limit
        Maximum number of messages passed to the logger.
    '''
    if logger is not None:
        # pylint: disable=protected-access
        c_handle = _ffi.new_handle(logger)
        c_cb = _lib.pyclingo_logger_callback
    else:
        c_handle = _ffi.NULL
        c_cb = _ffi.NULL
    return Symbol(_c_call('clingo_symbol_t', _lib.clingo_parse_term, string.encode(), c_cb, c_handle, message_limit))



'''
Tests for the backend/observer.
'''

from unittest import TestCase
from typing import Sequence, Tuple
from clingo import Control, Function, HeuristicType, Observer, Symbol, TruthValue

class TestObserverBackend(Observer):
    '''
    Test Observer.
    '''
    def __init__(self, case):
        self._case = case
        self.called = set()

    def init_program(self, incremental: bool) -> None:
        self.called.add('init_program')

    def begin_step(self) -> None:
        self.called.add('begin_step')

    def rule(self, choice: bool, head: Sequence[int], body: Sequence[int]) -> None:
        self.called.add('rule')
        self._case.assertTrue(choice)
        self._case.assertEqual(head, [1])
        self._case.assertEqual(body, [2, 3])

    def weight_rule(self, choice: bool, head: Sequence[int], lower_bound: int,
                    body: Sequence[Tuple[int,int]]) -> None:
        self.called.add('weight_rule')
        self._case.assertFalse(choice)
        self._case.assertEqual(head, [2])
        self._case.assertEqual(lower_bound, 1)
        self._case.assertEqual(body, [(2, 3), (4, 5)])

    def minimize(self, priority: int, literals: Sequence[Tuple[int,int]]) -> None:
        self.called.add('minimize')
        self._case.assertEqual(priority, 0)
        self._case.assertEqual(literals, [(2, 3), (4, 5)])

    def project(self, atoms: Sequence[int]) -> None:
        self.called.add('project')
        self._case.assertEqual(atoms, [2, 4])

    def output_atom(self, symbol: Symbol, atom: int) -> None:
        self.called.add('output_atom')
        self._case.assertEqual(symbol, Function('a'))
        self._case.assertEqual(atom, 2)

    def external(self, atom: int, value: TruthValue) -> None:
        self.called.add('external')
        self._case.assertEqual(atom, 3)
        self._case.assertEqual(value, TruthValue.Release)

    def assume(self, literals: Sequence[int]) -> None:
        self.called.add('assume')
        self._case.assertEqual(literals, [2, 3])

    def heuristic(self, atom: int, type_: HeuristicType, bias: int,
                  priority: int, condition: Sequence[int]) -> None:
        self.called.add('heuristic')
        self._case.assertEqual(atom, 2)
        self._case.assertEqual(type_, HeuristicType.Level)
        self._case.assertEqual(bias, 5)
        self._case.assertEqual(priority, 7)
        self._case.assertEqual(condition, [1, 3])

    def acyc_edge(self, node_u: int, node_v: int,
                  condition: Sequence[int]) -> None:
        self.called.add('acyc_edge')
        self._case.assertEqual(node_u, 1)
        self._case.assertEqual(node_v, 2)
        self._case.assertEqual(condition, [3, 4])

    def end_step(self) -> None:
        self.called.add('end_step')

class TestObserverTheory(Observer):
    '''
    Test Observer.
    '''
    def __init__(self, case):
        self._case = case
        self.called = set()

    def output_term(self, symbol: Symbol, condition: Sequence[int]) -> None:
        self.called.add('output_term')
        self._case.assertEqual(symbol, Function('t'))
        self._case.assertGreaterEqual(len(condition), 1)

    def theory_term_number(self, term_id: int, number: int) -> None:
        self.called.add('theory_term_number')
        self._case.assertEqual(number, 1)

    def theory_term_string(self, term_id : int, name : str) -> None:
        self.called.add('theory_term_string')
        self._case.assertEqual(name, "a")

    def theory_term_compound(self, term_id: int, name_id_or_type: int,
                             arguments: Sequence[int]) -> None:
        self.called.add('theory_term_compound')
        self._case.assertEqual(name_id_or_type, -1)
        self._case.assertGreaterEqual(len(arguments), 2)

    def theory_element(self, element_id: int, terms: Sequence[int],
                       condition: Sequence[int]) -> None:
        self.called.add('theory_element')
        self._case.assertEqual(len(terms), 1)
        self._case.assertEqual(len(condition), 2)

    def theory_atom(self, atom_id_or_zero: int, term_id: int,
                    elements: Sequence[int]) -> None:
        self.called.add('theory_atom')
        self._case.assertEqual(len(elements), 1)

class TestObserverTheoryWithGuard(Observer):
    '''
    Test Observer.
    '''
    def __init__(self, case):
        self._case = case
        self.called = set()

    def theory_term_string(self, term_id : int, name : str) -> None:
        self.called.add(f'theory_term_string: {name}')

    def theory_atom_with_guard(self, atom_id_or_zero: int, term_id: int,
                               elements: Sequence[int], operator_id: int,
                               right_hand_side_id: int) -> None:
        self.called.add('theory_atom_with_guard')
        self._case.assertEqual(len(elements), 0)

class TestBackend(TestCase):
    '''
    Tests basic solving and related functions.
    '''
    def test_backend(self):
        '''
        Test backend via observer.
        '''
        ctl = Control()
        obs = TestObserverBackend(self)
        ctl.register_observer(obs)
        with ctl.backend() as backend:
            self.assertIn('init_program', obs.called)
            self.assertIn('begin_step', obs.called)
            backend.add_atom()
            backend.add_atom(Function('a'))
            backend.add_rule([1], [2, 3], True)
            self.assertIn('rule', obs.called)
            backend.add_weight_rule([2], 1, [(2, 3), (4, 5)])
            self.assertIn('weight_rule', obs.called)
            backend.add_minimize(0, [(2, 3), (4, 5)])
            self.assertIn('minimize', obs.called)
            backend.add_project([2, 4])
            self.assertIn('project', obs.called)
            backend.add_heuristic(2, HeuristicType.Level, 5, 7, [1, 3])
            self.assertIn('heuristic', obs.called)
            backend.add_assume([2, 3])
            self.assertIn('assume', obs.called)
            backend.add_acyc_edge(1, 2, [3, 4])
            self.assertIn('acyc_edge', obs.called)
            backend.add_external(3, TruthValue.Release)
            self.assertIn('external', obs.called)
        self.assertIn('output_atom', obs.called)
        ctl.solve()
        self.assertIn('end_step', obs.called)

    def test_theory(self):
        '''
        Test observer via grounding.
        '''
        ctl = Control()
        obs = TestObserverTheory(self)
        ctl.register_observer(obs)
        ctl.add('base', [], '''\
        #theory test {
            t { };
            &a/0 : t, head
        }.
        {a; b}.
        #show t : a, b.
        &a { (1,a): a,b }.
        ''')
        ctl.ground([('base', [])])
        self.assertIn('output_term', obs.called)
        self.assertIn('theory_term_number', obs.called)
        self.assertIn('theory_term_string', obs.called)
        self.assertIn('theory_term_compound', obs.called)
        self.assertIn('theory_element', obs.called)
        self.assertIn('theory_atom', obs.called)
        ctl.solve()

    def test_theory_with_guard(self):
        '''
        Test observer via grounding.
        '''
        ctl = Control()
        obs = TestObserverTheoryWithGuard(self)
        ctl.register_observer(obs)
        ctl.add('base', [], '''\
        #theory test {
            t { };
            &a/0 : t, {=}, t, head
        }.
        &a { } = a.
        ''')
        ctl.ground([('base', [])])
        self.assertIn('theory_term_string: a', obs.called)
        self.assertIn('theory_term_string: =', obs.called)
        self.assertIn('theory_atom_with_guard', obs.called)
        ctl.solve()



'''
Tests for solving.
'''
from unittest import TestCase
from typing import cast
from clingo import Control, Function, Model, ModelType, SolveHandle, SolveResult, SymbolicAtom

from . util import _MCB, _check_sat, _p

class TestSolving(TestCase):
    '''
    Tests basic solving and related functions.
    '''

    def setUp(self):
        self.mcb = _MCB()
        self.mit = _MCB()
        self.ctl = Control(['0'])

    def tearDown(self):
        self.mcb = None
        self.mit = None
        self.ctl = None

    def test_solve_result_str(self):
        '''
        Test string representation of solve results.
        '''
        ret = self.ctl.solve()
        self.assertEqual(str(ret), 'SAT')
        self.assertRegex(repr(ret), 'SolveResult(.*)')

    def test_model_str(self):
        '''
        Test string representation of models.
        '''
        self.ctl.add('base', [], 'a.')
        self.ctl.ground([('base', [])])
        with cast(SolveHandle, self.ctl.solve(yield_=True)) as hnd:
            for mdl in hnd:
                self.assertEqual(str(mdl), "a")
                self.assertRegex(repr(mdl), "Model(.*)")

    def test_solve_cb(self):
        '''
        Test solving using callback.
        '''
        self.ctl.add("base", [], "1 {a; b} 1. c.")
        self.ctl.ground([("base", [])])
        _check_sat(self, cast(SolveResult, self.ctl.solve(on_model=self.mcb.on_model, yield_=False, async_=False)))
        self.assertEqual(self.mcb.models, _p(['a', 'c'], ['b', 'c']))
        self.assertEqual(self.mcb.last[0], ModelType.StableModel)

    def test_solve_async(self):
        '''
        Test asynchonous solving.
        '''
        self.ctl.add("base", [], "1 {a; b} 1. c.")
        self.ctl.ground([("base", [])])
        with cast(SolveHandle, self.ctl.solve(on_model=self.mcb.on_model, yield_=False, async_=True)) as hnd:
            _check_sat(self, hnd.get())
            self.assertEqual(self.mcb.models, _p(['a', 'c'], ['b', 'c']))

    def test_solve_yield(self):
        '''
        Test solving yielding models.
        '''
        self.ctl.add("base", [], "1 {a; b} 1. c.")
        self.ctl.ground([("base", [])])
        with cast(SolveHandle, self.ctl.solve(on_model=self.mcb.on_model, yield_=True, async_=False)) as hnd:
            for m in hnd:
                self.mit.on_model(m)
            _check_sat(self, hnd.get())
            self.assertEqual(self.mcb.models, _p(['a', 'c'], ['b', 'c']))
            self.assertEqual(self.mit.models, _p(['a', 'c'], ['b', 'c']))

    def test_solve_async_yield(self):
        '''
        Test solving yielding models asynchronously.
        '''
        self.ctl.add("base", [], "1 {a; b} 1. c.")
        self.ctl.ground([("base", [])])
        with cast(SolveHandle, self.ctl.solve(on_model=self.mcb.on_model, yield_=True, async_=True)) as hnd:
            while True:
                hnd.resume()
                _ = hnd.wait()
                m = hnd.model()
                if m is None:
                    break
                self.mit.on_model(m)
            _check_sat(self, hnd.get())
            self.assertEqual(self.mcb.models, _p(['a', 'c'], ['b', 'c']))
            self.assertEqual(self.mit.models, _p(['a', 'c'], ['b', 'c']))

    def test_solve_interrupt(self):
        '''
        Test interrupting solving.
        '''
        self.ctl.add("base", [], "1 { p(P,H): H=1..99 } 1 :- P=1..100.\n1 { p(P,H): P=1..100 } 1 :- H=1..99.")
        self.ctl.ground([("base", [])])
        with cast(SolveHandle, self.ctl.solve(async_=True)) as hnd:
            hnd.resume()
            hnd.cancel()
            ret = hnd.get()
            self.assertTrue(ret.interrupted)

        with cast(SolveHandle, self.ctl.solve(async_=True)) as hnd:
            hnd.resume()
            self.ctl.interrupt()
            ret = hnd.get()
            self.assertTrue(ret.interrupted)

    def test_solve_core(self):
        '''
        Test core retrieval.
        '''
        self.ctl.add("base", [], "3 { p(1..10) } 3.")
        self.ctl.ground([("base", [])])
        ass = []
        for atom in self.ctl.symbolic_atoms.by_signature("p", 1):
            ass.append(-atom.literal)
        ret = cast(SolveResult, self.ctl.solve(on_core=self.mcb.on_core, assumptions=ass))
        self.assertTrue(ret.unsatisfiable)
        self.assertTrue(len(self.mcb.core) > 7)

    def test_enum(self):
        '''
        Test core retrieval.
        '''
        self.ctl = Control(['0', '-e', 'cautious'])
        self.ctl.add("base", [], "1 {a; b} 1. c.")
        self.ctl.ground([("base", [])])
        self.ctl.solve(on_model=self.mcb.on_model)
        self.assertEqual(self.mcb.last[0], ModelType.CautiousConsequences)
        self.assertEqual([self.mcb.last[1]], _p(['c']))

        self.ctl = Control(['0', '-e', 'brave'])
        self.ctl.add("base", [], "1 {a; b} 1. c.")
        self.ctl.ground([("base", [])])
        self.ctl.solve(on_model=self.mcb.on_model)
        self.assertEqual(self.mcb.last[0], ModelType.BraveConsequences)
        self.assertEqual([self.mcb.last[1]], _p(['a', 'b', 'c']))

    def test_model(self):
        '''
        Test functions of model.
        '''
        def on_model(m: Model):
            self.assertTrue(m.contains(Function('a')))
            self.assertTrue(m.is_true(cast(SymbolicAtom, m.context.symbolic_atoms[Function('a')]).literal))
            self.assertFalse(m.is_true(1000))
            self.assertEqual(m.thread_id, 0)
            self.assertEqual(m.number, 1)
            self.assertFalse(m.optimality_proven)
            self.assertEqual(m.cost, [3])
            m.extend([Function('e')])
            self.assertSequenceEqual(m.symbols(theory=True), [Function('e')])
        self.ctl.add("base", [], "a. b. c. #minimize { 1,a:a; 1,b:b; 1,c:c }.")
        self.ctl.ground([("base", [])])
        self.ctl.solve(on_model=on_model)

    def test_control_clause(self):
        '''
        Test adding clauses while solving.
        '''
        self.ctl.add("base", [], "1 {a; b; c} 1.")
        self.ctl.ground([("base", [])])
        with cast(SolveHandle, self.ctl.solve(on_model=self.mcb.on_model, yield_=True, async_=False)) as hnd:
            for m in hnd:
                clause = []
                if m.contains(Function('a')):
                    clause.append((Function('b'), False))
                else:
                    clause.append((Function('a'), False))
                m.context.add_clause(clause)

            _check_sat(self, hnd.get())
            self.assertEqual(len(self.mcb.models), 2)

    def test_control_nogood(self):
        '''
        Test adding nogoods while solving.
        '''
        self.ctl.add("base", [], "1 {a; b; c} 1.")
        self.ctl.ground([("base", [])])
        with cast(SolveHandle, self.ctl.solve(on_model=self.mcb.on_model, yield_=True, async_=False)) as hnd:
            for m in hnd:
                clause = []
                if m.contains(Function('a')):
                    clause.append((Function('b'), True))
                else:
                    clause.append((Function('a'), True))
                m.context.add_nogood(clause)

            _check_sat(self, hnd.get())
            self.assertEqual(len(self.mcb.models), 2)



'''
Tests control.
'''
from unittest import TestCase
from typing import cast
from clingo import Control, Function, Number, SolveResult

class TestError(Exception):
    '''
    Test exception.
    '''

class Context:
    '''
    Simple context with some test functions.
    '''
    def cb_num(self, c):
        '''
        Simple test callback.
        '''
        return [Number(c.number + 1), Number(c.number - 1)]

    def cb_error(self):
        '''
        Simple test raising an error.
        '''
        raise TestError('test')

class TestControl(TestCase):
    '''
    Tests basic functions of the control object.
    '''
    def test_ground(self):
        '''
        Test grounding with context and parameters.
        '''
        ctx = Context()
        ctl = Control()
        ctl.add('part', ['c'], 'p(@cb_num(c)).')
        ctl.ground([('part', [Number(1)])], ctx)
        symbols = [atom.symbol for atom in ctl.symbolic_atoms]
        self.assertEqual(sorted(symbols), [Function('p', [Number(0)]), Function('p', [Number(2)])])

    def test_ground_error(self):
        '''
        Test grounding with context and parameters.
        '''
        ctx = Context()
        ctl = Control()
        ctl.add('part', ['c'], 'p(@cb_error()).')
        self.assertRaisesRegex(TestError, 'test', ctl.ground, [('part', [Number(1)])], ctx)

    def test_lower(self):
        '''
        Test lower bounds reported during optimization.
        '''
        ctl = Control(['--opt-str=usc,oll,0', '--stats=2'])
        ctl.add('base', [], '1 { p(X); q(X) } 1 :- X=1..3. #minimize { 1,p,X: p(X); 1,q,X: q(X) }.')
        ctl.ground([('base', [])])
        lower = []
        self.assertTrue(cast(SolveResult, ctl.solve(on_unsat=lower.append)).satisfiable)
        self.assertEqual(lower, [[1], [2], [3]])
        self.assertEqual(ctl.statistics['summary']['lower'], [3.0])

    def test_error_handling(self):
        '''
        Test basic error handling during solving.
        '''
        ctl = Control()
        ctl.add('base', [], '1 {a; b} 1.')
        ctl.ground([('base', [])])

        self.assertRaises(ZeroDivisionError, lambda: ctl.solve(on_model=lambda m: 1 / 0))

        with ctl.solve(on_model=lambda m: 1 / 0, yield_=True) as handle:
            self.assertRaises(ZeroDivisionError, lambda: [_ for _ in handle])

        # Note: currently clasp does not store and re-raise the exception in
        # asynchronous mode, so we get a runtime error instead
        with ctl.solve(on_model=lambda m: 1 / 0, async_=True) as handle:
            self.assertRaises(RuntimeError, handle.get)



'''
Functions and classes related to statistics.

Examples
--------
The following example shows how to add custom statistics and dump the
statistics in json format:

    >>> from json import dumps
    >>> from clingo.control import Control
    >>>
    >>> def on_statistics(step, accu):
    ...     accu["example"] = 42
    ...
    >>> ctl = Control(['--stats'])
    >>> ctl.add("base", [], "{a}.")
    >>> ctl.ground([("base", [])])
    >>> print(ctl.solve(on_statistics=on_statistics))
    SAT
    >>> print(dumps(ctl.statistics['user_accu'], sort_keys=True,
    ...             indent=4, separators=(',', ': ')))
    {
        "example": 42.0
    }
    >>> print(dumps(ctl.statistics['summary']['times'], sort_keys=True,
    ...             indent=4, separators=(',', ': ')))
    {
        "cpu": 0.000785999999999995,
        "sat": 7.867813110351562e-06,
        "solve": 2.288818359375e-05,
        "total": 0.0007848739624023438,
        "unsat": 0.0
    }
'''
from typing import (AbstractSet, Any, Iterable, KeysView, Sequence, Tuple, MutableMapping, MutableSequence, Union,
                    ValuesView,
                    cast)
from collections import abc

from ._internal import _c_call, _handle_error, _lib, _to_str

__all__ = [ 'StatisticsArray', 'StatisticsMap', 'StatisticsValue' ]

StatisticsValue = Union['StatisticsArray', 'StatisticsMap', float]

def _statistics_type(stats, key):
    return _c_call('clingo_statistics_type_t', _lib.clingo_statistics_type, stats, key)

def _statistics(stats, key):
    '''
    Transform clingo's statistics into python type.
    '''
    type_ = _statistics_type(stats, key)

    if type_ == _lib.clingo_statistics_type_value:
        return _c_call('double', _lib.clingo_statistics_value_get, stats, key)

    if type_ == _lib.clingo_statistics_type_array:
        ret = []
        for i in range(_c_call('size_t', _lib.clingo_statistics_array_size, stats, key)):
            ret.append(_statistics(stats, _c_call('uint64_t', _lib.clingo_statistics_array_at, stats, key, i)))
        return ret

    assert type_ == _lib.clingo_statistics_type_map
    ret = {}
    for i in range(_c_call('size_t', _lib.clingo_statistics_map_size, stats, key)):
        name = _c_call('char*', _lib.clingo_statistics_map_subkey_name, stats, key, i)
        subkey = _c_call('uint64_t', _lib.clingo_statistics_map_at, stats, key, name)
        ret[_to_str(name)] = _statistics(stats, subkey)
    return ret

def _mutable_statistics_value_type(value):
    if isinstance(value, str):
        raise TypeError('unexpected string')
    if isinstance(value, abc.Sequence):
        return _lib.clingo_statistics_type_array
    if isinstance(value, abc.Mapping):
        return _lib.clingo_statistics_type_map
    return _lib.clingo_statistics_type_value

def _mutable_statistics(stats) -> 'StatisticsMap':
    key = _c_call('uint64_t', _lib.clingo_statistics_root, stats)
    return cast(StatisticsMap, _mutable_statistics_get(stats, key))

def _mutable_statistics_get(stats, key) -> StatisticsValue:
    type_ = _statistics_type(stats, key)

    if type_ == _lib.clingo_statistics_type_array:
        return StatisticsArray(stats, key)

    if type_ == _lib.clingo_statistics_type_map:
        return StatisticsMap(stats, key)

    assert type_ == _lib.clingo_statistics_type_value
    return _c_call('double', _lib.clingo_statistics_value_get, stats, key)

def _mutable_statistics_set(stats, key, type_, value, update):
    if type_ == _lib.clingo_statistics_type_map:
        StatisticsMap(stats, key).update(value)
    elif type_ == _lib.clingo_statistics_type_array:
        StatisticsArray(stats, key).update(value)
    else:
        assert type_ == _lib.clingo_statistics_type_value
        if callable(value):
            if update:
                uval = value(_c_call('double', _lib.clingo_statistics_value_get, stats, key))
            else:
                uval = value(None)
        else:
            uval = value
        _handle_error(_lib.clingo_statistics_value_set(stats, key, uval))

class StatisticsArray(MutableSequence[StatisticsValue]):
    '''
    Class to modify statistics stored in an array.

    Notes
    -----
    Only inplace concatenation and no deletion is supported.
    '''
    def __init__(self, rep, key):
        self._rep = rep
        self._key = key

    def __len__(self):
        return _c_call('size_t', _lib.clingo_statistics_array_size, self._rep, self._key)

    def __getitem__(self, index):
        if index < 0 or index >= len(self):
            raise IndexError('invalid index')
        key = _c_call('uint64_t', _lib.clingo_statistics_array_at, self._rep, self._key, index)
        return _mutable_statistics_get(self._rep, key)

    def __setitem__(self, index, value):
        key = _c_call('uint64_t', _lib.clingo_statistics_array_at, self._rep, self._key, index)
        type_ = _statistics_type(self._rep, key)
        _mutable_statistics_set(self._rep, key, type_, value, True)

    def insert(self, index, value):
        '''
        This method is not supported.
        '''
        raise NotImplementedError('insertion is not supported')

    def __delitem__(self, index):
        raise NotImplementedError('deletion is not supported')

    def __add__(self, other):
        raise NotImplementedError('only inplace concatenation is supported')

    def __iadd__(self, other):
        for x in other:
            self.append(x)
        return self

    def append(self, value: Any) -> None:
        '''
        Append a value.

        Parameters
        ----------
        value
            A nested structure composed of floats, sequences, and mappings.
        '''
        type_ = _mutable_statistics_value_type(value)
        key = _c_call('uint64_t', _lib.clingo_statistics_array_push, self._rep, self._key, type_)
        _mutable_statistics_set(self._rep, key, type_, value, False)

    def extend(self, values: Iterable[Any]) -> None:
        '''
        Extend the statistics array with the given values.

        Paremeters
        ----------
        values
            A sequence of nested structures composed of floats, sequences, and
            mappings.
        '''
        self += values

    def update(self, values: Sequence[Any]) -> None:
        '''
        Update a statistics array.

        Parameters
        ----------
        values
            A sequence of nested structures composed of floats, callable,
            sequences, and mappings. A callable can be used to update an
            existing value, it receives the previous numeric value (or None if
            absent) as argument and must return an updated numeric value.
        '''
        n = len(self)
        for idx, val in enumerate(values):
            if idx < n:
                self[idx] = val
            else:
                self.append(val)

class StatisticsMap(MutableMapping[str, StatisticsValue]):
    '''
    Object to capture statistics stored in a map.

    This class does not support item deletion.
    '''
    def __init__(self, rep, key):
        self._rep = rep
        self._key = key

    def __len__(self):
        return _c_call('size_t', _lib.clingo_statistics_map_size, self._rep, self._key)

    def __getitem__(self, name: str):
        if not name in self:
            raise KeyError('key not found')
        key = _c_call('uint64_t', _lib.clingo_statistics_map_at, self._rep, self._key, name.encode())
        return _mutable_statistics_get(self._rep, key)

    def __setitem__(self, name: str, value: Any):
        has_key = name in self
        if has_key:
            key = _c_call('uint64_t', _lib.clingo_statistics_map_at, self._rep, self._key, name.encode())
            type_ = _statistics_type(self._rep, key)
        else:
            type_ = _mutable_statistics_value_type(value)
            key = _c_call('uint64_t', _lib.clingo_statistics_map_add_subkey, self._rep, self._key, name.encode(), type_)
        _mutable_statistics_set(self._rep, key, type_, value, has_key)

    def __delitem__(self, index):
        raise NotImplementedError('deletion is not supported')

    def __contains__(self, name):
        # note: has to be done like this because of python's container model
        if not isinstance(name, str):
            return False
        return _c_call('bool', _lib.clingo_statistics_map_has_subkey, self._rep, self._key, name.encode())

    def __iter__(self):
        return iter(self.keys())

    def items(self) -> AbstractSet[Tuple[str, StatisticsValue]]:
        '''
        Return the items of the map.
        '''
        ret = []
        for i in range(len(self)):
            name = _c_call('char*', _lib.clingo_statistics_map_subkey_name, self._rep, self._key, i)
            key = _c_call('uint64_t', _lib.clingo_statistics_map_at, self._rep, self._key, name)
            ret.append((_to_str(name), _mutable_statistics_get(self._rep, key)))
        # Note: to shut up the type checker; this should work fine in practice
        return cast(AbstractSet[Tuple[str, StatisticsValue]], ret)

    def keys(self) -> KeysView[str]:
        '''
        Return the keys of the map.
        '''
        ret = []
        for i in range(len(self)):
            ret.append(_to_str(_c_call('char*', _lib.clingo_statistics_map_subkey_name, self._rep, self._key, i)))
        # Note: to shut up the type checker; this should work fine in practice
        return cast(KeysView[str], ret)

    def update(self, values):
        '''
        Update the map with the given values.

        Parameters
        ----------
        values
            A mapping of nested structures composed of floats, callable,
            sequences, and mappings. A callable can be used to update an
            existing value, it receives the previous numeric value (or None if
            absent) as argument and must return an updated numeric value.
        '''
        for key, value in values.items():
            self[key] = value

    def values(self) -> ValuesView[StatisticsValue]:
        '''
        Return the values of the map.
        '''
        ret = []
        for i in range(len(self)):
            name = _c_call('char*', _lib.clingo_statistics_map_subkey_name, self._rep, self._key, i)
            key = _c_call('uint64_t', _lib.clingo_statistics_map_at, self._rep, self._key, name)
            ret.append(_mutable_statistics_get(self._rep, key))
        # Note: to shut up the type checker; this should work fine in practice
        return cast(ValuesView[StatisticsValue], ret)
